import abc
import inspect
import types
import six

from smqtk.utils import merge_dict


@six.add_metaclass(abc.ABCMeta)
class Configurable (object):
    """
    Interface for objects that should be configurable via a configuration
    dictionary consisting of JSON types.
    """

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        turning those argument names into configuration dictionary keys. If any
        of those arguments have defaults, we will add those values into the
        configuration dictionary appropriately. The dictionary returned should
        only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        >>> class SimpleConfig(Configurable):
        ...     def __init__(self, a=1, b='foo'):
        ...         self.a = a
        ...         self.b = b
        ...     def get_config(self):
        ...         return {'a': self.a, 'b': self.b}
        >>> self = SimpleConfig()
        >>> config = self.get_default_config()
        >>> assert config == {'a': 1, 'b': 'foo'}
        """
        if isinstance(cls.__init__, (types.MethodType, types.FunctionType)):
            argspec = inspect.getargspec(cls.__init__)

            # Ignores potential *args or **kwargs present
            params = argspec.args[1:]  # skipping ``self`` arg
            num_params = len(params)

            if argspec.defaults:
                num_defaults = len(argspec.defaults)
                vals = ((None,) * (num_params - num_defaults)
                        + argspec.defaults)
            else:
                vals = (None,) * num_params

            return dict(zip(params, vals))

        # No constructor explicitly defined on this class
        return {}

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless an instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: Configurable

        """
        # The simple case is that the class doesn't require any special
        # parameters other than those that can be provided via the JSON
        # specification, which we cover here. If an implementation needs
        # something more special, they can override this function.
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        return cls(**config_dict)

    @abc.abstractmethod
    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """



"""
Top level flask application
"""

import base64
import json
import os
import os.path
import threading

import flask
from flask_cors import cross_origin
import six
from werkzeug.exceptions import NotFound
from werkzeug.wsgi import peek_path_info, pop_path_info

from smqtk.utils import DatabaseInfo
from smqtk.utils import merge_dict
from smqtk.utils.mongo_sessions import MongoSessionInterface
from smqtk.web import SmqtkWebApp

from .modules.login import LoginMod
from .modules.iqr import IqrSearch


SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


def generate_csrf_token():
    """
    Create a random string token for CSRF protection.

    Uses ``os.urandom`` so this should be sufficient for cryptographic uses.

    """
    BYTES = 64
    return base64.b64encode(os.urandom(BYTES)).decode('utf-8')


class IqrSearchDispatcher (SmqtkWebApp):
    """
    Application that dispatches to IQR application instances per sub-path.  We
    can be seeded with a set amount of instances with the ``iqr_tabs``
    configuration section, which consists of a prefix key (what would be used in
    the URL) to the configuration for that instance.  A ``__default__`` is
    provided upon configuration generation to act as a template.  The
    ``__default__`` value is ignored at runtime.

    New IQR instances can be dynamically added via a POST to the url root
    (``/``).
    """

    # Prefixes that ignore dispatch to IQR application instances
    PREFIX_BLACKLIST = {
        'static',
        'login',
        'login.passwd',
        'logout',
    }

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        c = super(IqrSearchDispatcher, cls).get_default_config()
        merge_dict(c, {
            "mongo": {
                "server": "127.0.0.1:27017",
                "database": "smqtk",
            },
            # Each entry in this mapping generates a new tab in the GUI
            "iqr_tabs": {
                "__default__": IqrSearch.get_default_config(),
            },
        })
        return c

    def __init__(self, json_config):
        super(IqrSearchDispatcher, self).__init__(json_config)

        #
        # Database setup using Mongo
        #
        h, p = self.json_config['mongo']['server'].split(':')
        n = self.json_config['mongo']['database']
        self.db_info = DatabaseInfo(h, p, n)

        # Use mongo for session storage.
        # -> This allows session modification during Flask methods called from
        #    AJAX routines (default Flask sessions do not)
        self.session_interface = MongoSessionInterface(self.db_info.host,
                                                       self.db_info.port,
                                                       self.db_info.name)

        #
        # Misc. Setup
        #

        # Add 'do' statement usage
        self.jinja_env.add_extension('jinja2.ext.do')

        #
        # Modules
        #
        # Load up required and optional module blueprints
        #

        # Mapping of IqrSearch application instances from their ID string
        #: :type: dict[str, IqrSearch]
        self.instances = {}
        self.instances_lock = threading.Lock()

        # Login module
        self._log.info("Initializing Login Blueprint")
        self.module_login = LoginMod('login', self)
        self.register_blueprint(self.module_login)

        # IQR modules
        # - for each entry in 'iqr_tabs', initialize a separate IqrSearch
        #   instance.
        for tab_name, tab_config in six.iteritems(self.json_config['iqr_tabs']):
            if tab_name == "__default__":
                # skipping default config sample
                continue
            self._log.info("Initializing IQR instance '%s'", tab_name)
            self.init_iqr_app(tab_config, tab_name)

        #
        # Basic routing
        #

        @self.route('/', methods=['GET'])
        def index():
            # self._log.info("Session: %s", flask.session.items())
            # noinspection PyUnresolvedReferences
            return flask.render_template("index.html",
                                         instance_keys=list(self.instances.keys()),
                                         debug=self.debug)

        @self.route('/', methods=['POST'])
        @cross_origin(origins='*', vary_header=True)
        @self.module_login.login_required
        def add_instance():
            """
            Initialize new IQR instance given an ID for that instance, and the
            configuration for it.
            """
            # TODO: Something where only user that created instance can access
            #       it?
            prefix = flask.request.form['prefix']
            config = json.loads(flask.request.form['config'])

            # the URL prefix of the new IqrSearch instance
            new_url = flask.request.host_url + prefix
            self._log.info("New URL with route: %s", new_url)

            self.init_iqr_app(config, prefix)

            return flask.jsonify({
                'prefix': prefix,
                'url': new_url,
            })

    @staticmethod
    def _apply_csrf_protect(app):
        # Establish CSRF protection

        # Synchronized keys also defined in ``/static/js/smqtk.vars.js``.
        # - Session key is also what's used in forms
        CSRF_FORM_KEY = CSRF_SESSION_TOKEN_KEY = "_csrf_token"
        CSRF_HEADER_KEY = "X-Csrf-Token"

        # CSRF Protection
        @app.before_request
        def csrf_protect():
            if flask.request.method in ["POST", "PUT", "DELETE"]:
                session_token = flask.session.get(CSRF_SESSION_TOKEN_KEY, None)

                # CSRF token will either be in header or form, but must be
                # present in one of them.
                req_header_token = flask.request.headers.get(CSRF_HEADER_KEY,
                                                             None)
                req_form_token = flask.request.form.get(CSRF_FORM_KEY, None)

                if not session_token or session_token != (req_header_token or
                                                          req_form_token):
                    flask.abort(400)

        def get_csrf_session_token():
            """
            Initialize and return a specific, secure token for CSRF mitigation
            per session.
            """
            # Initialize CSRF protection token in session if not there.
            if CSRF_SESSION_TOKEN_KEY not in flask.session:
                flask.session[CSRF_SESSION_TOKEN_KEY] = generate_csrf_token()
            return flask.session['_csrf_token']

        # Add CSRF templating helper token generator
        app.jinja_env.globals['csrf_form_key'] = CSRF_FORM_KEY
        app.jinja_env.globals['csrf_token_gen'] = get_csrf_session_token

        return app

    def init_iqr_app(self, config, prefix):
        """
        Initialize IQR sub-application given a configuration for it and a prefix

        :param config: IqrSearch plugin configuration dictionary
        :type config: dict

        :param prefix: URL prefix for the instance
        :type prefix: str

        :return: Application instance.
        :rtype: IqrSearch

        """
        with self.instances_lock:
            if prefix not in self.instances:
                self._log.info("Initializing IQR instance '%s'", prefix)
                self._log.debug("IQR tab config:\n%s", config)
                # Strip any keys that are not expected by IqrSearch
                # constructor
                expected_keys = list(IqrSearch.get_default_config().keys())
                for k in set(config).difference(expected_keys):
                    self._log.debug("Removing unexpected key: %s", k)
                    del config[k]
                self._log.debug("Base app config: %s", self.config)

                a = IqrSearch.from_config(config, self)
                a.config.update(self.config)
                a.secret_key = self.secret_key
                a.session_interface = self.session_interface
                a.jinja_env.add_extension('jinja2.ext.do')
                self._apply_csrf_protect(a)

                self.instances[prefix] = a
            else:
                self._log.debug("Existing IQR instance for prefix: '%s'",
                                prefix)
                a = self.instances[prefix]

        return a

    def get_application(self, prefix):
        """
        Get the application for the given ``prefix`` or the NotFound exception
        if an application does not yet exist for the ``prefix``.

        :param prefix: Prefix name of the IQR application instance
        :type prefix: str

        :return: Application instance or None if there is no instance for the
            given ``prefix``.
        :rtype: IqrSearch | None

        """
        with self.instances_lock:
            return self.instances.get(prefix, None)

    def __call__(self, environ, start_response):
        path_prefix = peek_path_info(environ)
        self._log.debug("Base application __call__ path prefix: '%s'",
                        path_prefix)

        if path_prefix and path_prefix not in self.PREFIX_BLACKLIST:
            app = self.get_application(path_prefix)
            if app is not None:
                pop_path_info(environ)
            else:
                self._log.debug("No IQR application registered for prefix: "
                                "'%s'", path_prefix)
                app = NotFound()
        else:
            self._log.debug("No prefix or prefix in blacklist. "
                            "Using dispatcher app.")
            app = self.wsgi_app

        return app(environ, start_response)

    def run(self, host=None, port=None, debug=False, **options):
        # Establish CSRF protection
        self._apply_csrf_protect(self)

        super(IqrSearchDispatcher, self).run(host, port, debug, **options)


APPLICATION_CLASS = IqrSearchDispatcher



import os
import os.path as osp
import re

from six.moves import cPickle as pickle

from smqtk.representation import DataElement, DataSet
from smqtk.utils import file_utils
from smqtk.utils.string_utils import partition_string


class DataFileSet (DataSet):
    """
    File-based data set

    This implementation serializes each stored DataElement to a single file on
    disk. The only memory overhead for this implementation is the mapping of
    data element original UUID to the file path on disk.

    File sets are initialized with a root directory, under which it finds and
    serializes DataElement instances. DataElement implementations are required
    to be picklable, so this is a valid assumption.

    Serializations are saved/loaded with respect to a ``uuid_chunk`` property,
    which defines how files are split into subdirectories based on their UUID.
    This is helpful when a lot of files are being stored as too many files in a
    single directory generally slows down filesystem access to that directory.

    """
    # TODO: Use file-based locking mechanism to make thread/process safe

    # Filename template for serialized files. Requires template
    SERIAL_FILE_TEMPLATE = "UUID_%s.dataElement"

    # Regex for matching file names as valid FileSet serialized elements
    # - yields two groups, the first is the UUID, the second is the SHA1 sum
    SERIAL_FILE_RE = re.compile("UUID_(\w+)\.dataElement")

    @classmethod
    def is_usable(cls):
        """
        Check whether this data set implementations is available for use.

        This is always true for this implementation as there are no required 3rd
        party dependencies

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        return True

    def __init__(self, root_directory, uuid_chunk=10, pickle_protocol=-1):
        """
        Initialize a new or existing file set from a root directory.

        If ``uuid_chunk`` is set to None or 1 we look for and save file
        caches in a flat manner under the root directory given. If greater than
        1, we split ``DataElement`` UUIDs into that many segments and use all
        but the last segment to define what sub-directory to place the
        serialization under the given ``root_directory``. For example, if
        ``uuid_chunk`` is 3 and a file's UUID is "abcdef", we store the
        element's serialization under the directory "<root_directory>/ab/cd/".
        We don't use the last segment in order to allow multiple files, but not
        all of them, to exist in leaf directories as opposed to there being a
        separate directory for each file, which is excessive.

        :param root_directory: Directory that this file set is based in. For
            relative path resolution, see the ``work_relative`` parameter
            description.
        :type root_directory: str

        :param uuid_chunk: Number of segments to split data element UUID
            into when saving element serializations. This should be None or a
            positive integer.
        :type uuid_chunk: None | int

        :param pickle_protocol: Pickling protocol to use. We will use -1 by
            default (latest version, probably binary).
        :type pickle_protocol: int

        """
        super(DataFileSet, self).__init__()
        if not (uuid_chunk is None or uuid_chunk > 0):
            raise ValueError('Uuid chunk must either be None or a positive '
                             'integer.')

        self._root_dir = os.path.expanduser(root_directory)
        self._uuid_chunk = uuid_chunk
        self.pickle_protocol = pickle_protocol

        self._log.debug("Initializing FileSet under root dir: %s",
                        self._root_dir)

    def _iter_file_tree(self):
        """
        Iterate over our file tree, yielding the file paths of serialized
        elements found in the expected sub-directories.
        """
        # Select how far we need to descent into root based on chunk level.
        recurse = (self._uuid_chunk and self._uuid_chunk - 1) or None

        for fp in file_utils.iter_directory_files(self._root_dir, recurse):
            m = self.SERIAL_FILE_RE.match(osp.basename(fp))
            if m:
                # Where file is under root to see if it is a file we care about
                # according to our ``uuid_chunk`` value.
                # - ``seg`` includes file name, so it will at least be of length
                #   1, so its unmodified length should equal ``uuid_chunk``.
                # - Exceptions: len(seg) == 1 and uuid_chunk in {None, 0, 1}
                seg = osp.relpath(fp, self._root_dir).split(os.sep)
                if self._uuid_chunk in {None, 1} and len(seg) == 1:
                    yield fp
                elif len(seg) == self._uuid_chunk:
                    yield fp

    def _containing_dir(self, uuid):
        """
        Return the containing directory for something with the given UUID value
        """
        if not self._uuid_chunk:
            # No sub-directory storage configured
            return self._root_dir

        str_uuid = str(uuid)
        # TODO(paul.tunison): Modify uuid string if to short for set UUID chunk.
        #     e.g. if uuid is the integer 1 and chunk size is 10, we should
        #     convert strigified result to be at least length 10?
        #     Do this in _fp_for_uuid method?
        leading_parts = partition_string(str_uuid, self._uuid_chunk)[:-1]
        return osp.join(self._root_dir, *leading_parts)

    def _fp_for_uuid(self, uuid):
        """
        Return the filepath to where an element with the given UUID would be
        saved.
        """
        return osp.join(self._containing_dir(uuid),
                        self.SERIAL_FILE_TEMPLATE % uuid)

    def __iter__(self):
        """
        :return: Generator over the DataElements contained in this set in no
            particular order.
        """
        for fp in self._iter_file_tree():
            # deserialize and yield
            with open(fp) as f:
                yield pickle.load(f)

    def get_config(self):
        return {
            "root_directory": self._root_dir,
            "uuid_chunk": self._uuid_chunk,
            "pickle_protocol": self.pickle_protocol,
        }

    def count(self):
        """
        :return: The number of data elements in this set.
        :rtype: int
        """
        c = 0
        for _ in self._iter_file_tree():
            c += 1
        return c

    def uuids(self):
        """
        :return: A new set of uuids represented in this data set.
        :rtype: set
        """
        s = set()
        for de in self:
            s.add(de.uuid())
        return s

    def has_uuid(self, uuid):
        """
        Test if the given uuid refers to an element in this data set.

        :param uuid: Unique ID to test for inclusion. This should match the type
            that the set implementation expects or cares about.

        :return: True if the given uuid matches an element in this set, or False
            if it does not.
        :rtype: bool

        """
        # Try to access the expected file path like a hash table
        return osp.isfile(self._fp_for_uuid(uuid))

    def add_data(self, *elems):
        """
        Add the given data element(s) instance to this data set.

        :param elems: Data element(s) to add
        :type elems: list[smqtk.representation.DataElement]

        """
        for e in elems:
            assert isinstance(e, DataElement), \
                "Not given a DataElement for addition: '%s'" % e
            uuid = str(e.uuid())
            fp = self._fp_for_uuid(uuid)
            file_utils.safe_create_dir(osp.dirname(fp))
            with open(fp, 'wb') as f:
                pickle.dump(e, f, self.pickle_protocol)
            self._log.debug("Wrote out element %s", e)

    def get_data(self, uuid):
        """
        Get the data element the given uuid references, or raise an
        exception if the uuid does not reference any element in this set.

        :raises KeyError: If the given uuid does not refer to an element in
            this data set.

        :param uuid: The uuid of the element to retrieve.

        :return: The data element instance for the given uuid.
        :rtype: smqtk.representation.DataElement

        """
        fp = self._fp_for_uuid(uuid)
        if not osp.isfile(fp):
            raise KeyError(uuid)
        else:
            with open(fp, 'rb') as f:
                return pickle.load(f)


DATA_SET_CLASS = DataFileSet



"""
Print out information about what plugins are currently usable and the
documentation headers for each implementation.
"""
from __future__ import print_function

import argparse
import json
import logging

import smqtk.algorithms
import smqtk.algorithms.nn_index.hash_index
import smqtk.algorithms.nn_index.lsh.functors
import smqtk.representation
import smqtk.utils.bin_utils
import smqtk.utils.plugin


def cli():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-v", "--verbose",
                        default=False, action="store_true",
                        help="Output additional debug logging.")
    parser.add_argument("--defaults",
                        default=False, type=str,
                        help="Optionally generate default configuration blocks "
                             "for each plugin structure and output as JSON to "
                             "the specified path.")
    return parser


def format_cls_description(name, cls):
    return "%s\n%s" % (name, cls)


def main():
    args = cli().parse_args()

    llevel = logging.INFO
    if args.verbose:
        llevel = logging.DEBUG
    smqtk.utils.bin_utils.initialize_logging(logging.getLogger("smqtk"), llevel)

    collect_defaults = args.defaults
    defaults = {}

    def collect_configs(name, impl_map):
        """
        :type name: str
        :type impl_map: dict
        """
        if collect_defaults:
            defaults[name] = smqtk.utils.plugin.make_config(impl_map)

    log = logging.getLogger("smqtk.checkPlugins")

    # Key is the interface type name
    plugin_info = {}
    # List of plugin_info keys in order they were added
    plugin_type_list = []

    def collect_plugins(type_name, impl_getter_fn):
        log.info("Checking %s plugins", type_name)
        plugin_type_list.append(type_name)
        impl_map = impl_getter_fn()
        plugin_info[plugin_type_list[-1]] = impl_map
        collect_configs(type_name, impl_map)

    #
    # smqtk.representation
    #
    collect_plugins('DataElement',
                    smqtk.representation.get_data_element_impls)
    collect_plugins('DataSet',
                    smqtk.representation.get_data_set_impls)
    collect_plugins('DescriptorElement',
                    smqtk.representation.get_descriptor_element_impls)
    collect_plugins('DescriptorIndex',
                    smqtk.representation.get_descriptor_index_impls)
    collect_plugins('KeyValueStore',
                    smqtk.representation.get_key_value_store_impls)

    #
    # smqtk.algorithms
    #
    collect_plugins('Classifier',
                    smqtk.algorithms.get_classifier_impls)
    collect_plugins('DescriptorGenerator',
                    smqtk.algorithms.get_descriptor_generator_impls)
    collect_plugins('HashIndex',
                    smqtk.algorithms.get_hash_index_impls)
    collect_plugins('LshFunctor',
                    smqtk.algorithms.get_lsh_functor_impls)
    collect_plugins('NearestNeighborIndex',
                    smqtk.algorithms.get_nn_index_impls)
    collect_plugins('RelevancyIndex',
                    smqtk.algorithms.get_relevancy_index_impls)

    #
    # Print-out
    #
    print()
    print()
    for k in plugin_type_list:
        print("[Type]", k)
        print('=' * (7 + len(k)))
        print()
        for l, t in plugin_info[k].items():
            print(":: " + l)
            if t.__doc__:
                print(t.__doc__.rstrip())
                print()
        print()
        print()

    if collect_defaults:
        with open(collect_defaults, 'w') as f:
            json.dump(defaults, f, indent=4, sort_keys=True,
                      separators=(',', ': '))
        log.info("Wrote default configuration dictionaries to: %s",
                 collect_defaults)


if __name__ == "__main__":
    main()



"""
Stub abstract class implementations.
"""
from smqtk.algorithms import Classifier, DescriptorGenerator, \
    NearestNeighborsIndex
from smqtk.representation import DescriptorIndex


STUB_MODULE_PATH = __name__


class StubDescriptorIndex (DescriptorIndex):
    @classmethod
    def is_usable(cls):
        return True

    def add_many_descriptors(self, descriptors):
        pass

    def count(self):
        pass

    def iteritems(self):
        pass

    def clear(self):
        pass

    def remove_descriptor(self, uuid):
        pass

    def remove_many_descriptors(self, uuids):
        pass

    def get_config(self):
        pass

    def get_many_descriptors(self, uuids):
        pass

    def has_descriptor(self, uuid):
        pass

    def iterkeys(self):
        pass

    def get_descriptor(self, uuid):
        pass

    def iterdescriptors(self):
        pass

    def add_descriptor(self, descriptor):
        pass


class StubClassifier (Classifier):
    """
    Classifier stub for testing IqrService.
    """

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    def get_labels(self):
        pass

    def _classify(self, d):
        pass


class StubDescrGenerator (DescriptorGenerator):
    """
    DescriptorGenerator stub for testing IqrService.
    """

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    def valid_content_types(self):
        pass

    def _compute_descriptor(self, data):
        pass


class StubNearestNeighborIndex (NearestNeighborsIndex):
    """
    NearestNeighborIndex stub for testing IqrService.
    """

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    def count(self):
        pass

    def _build_index(self, descriptors):
        pass

    def _update_index(self, descriptors):
        pass

    def _remove_from_index(self, uids):
        pass

    def _nn(self, d, n=1):
        pass



from __future__ import division, print_function
import mock
import numpy
import unittest

from smqtk.representation import DescriptorElement


class DummyDescriptorElement (DescriptorElement):

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        """ stub """

    def set_vector(self, new_vec):
        """ stub """

    def has_vector(self):
        """ stub """

    def vector(self):
        """ stub """


class TestDescriptorElementAbstract (unittest.TestCase):

    def test_init(self):
        expected_uuid = 'some uuid'
        expected_type_str = 'some type'
        de = DummyDescriptorElement(expected_type_str, expected_uuid)
        self.assertEqual(de.type(), expected_type_str)
        self.assertEqual(de.uuid(), expected_uuid)

    def test_equality(self):
        de1 = DummyDescriptorElement('t', 'u1')
        de2 = DummyDescriptorElement('t', 'u2')
        de1.vector = de2.vector = \
            mock.Mock(return_value=numpy.random.randint(0, 10, 10))

        self.assertTrue(de1 == de1)
        self.assertTrue(de2 == de2)
        self.assertTrue(de1 == de2)
        self.assertFalse(de1 != de2)

    def test_equality_diffTypeStr(self):
        # Type strings no longer considered in descriptor equality nor hashing.
        v = numpy.random.randint(0, 10, 10)
        d1 = DummyDescriptorElement('a', 'u')
        d2 = DummyDescriptorElement('b', 'u')
        d1.vector = d2.vector = mock.Mock(return_value=v)
        self.assertFalse(d1 != d2)
        self.assertTrue(d1 == d2)

    def test_nonEquality_diffInstance(self):
        # diff instance
        de = DummyDescriptorElement('a', 'b')
        self.assertFalse(de == 'string')
        self.assertTrue(de != 'string')

    def test_nonEquality_diffVectors(self):
        # different vectors (same size)
        v1 = numpy.random.randint(0, 10, 10)
        v2 = numpy.random.randint(0, 10, 10)

        d1 = DummyDescriptorElement('a', 'b')
        d1.vector = mock.Mock(return_value=v1)

        d2 = DummyDescriptorElement('a', 'b')
        d2.vector = mock.Mock(return_value=v2)

        self.assertFalse(d1 == d2)
        self.assertTrue(d1 != d2)

    def test_nonEquality_diffVectorSize(self):
        # different sized vectors
        v1 = numpy.random.randint(0, 10, 10)
        v2 = numpy.random.randint(0, 10, 100)

        d1 = DummyDescriptorElement('a', 'b')
        d1.vector = mock.Mock(return_value=v1)

        d2 = DummyDescriptorElement('a', 'b')
        d2.vector = mock.Mock(return_value=v2)

        self.assertFalse(d1 == d2)
        self.assertTrue(d1 != d2)

    def test_hash(self):
        # Hash of a descriptor element is solely based on the UUID value of that
        # element.
        t1 = 'a'
        uuid1 = 'some uuid'
        de1 = DummyDescriptorElement(t1, uuid1)

        t2 = 'b'
        uuid2 = 'some uuid'
        de2 = DummyDescriptorElement(t2, uuid2)

        self.assertEqual(hash(de1), hash(uuid1))
        self.assertEqual(hash(de2), hash(uuid2))
        self.assertEqual(hash(de1), hash(de2))

    def test_getstate(self):
        expected_type = 'a'
        expected_uid = 'b'
        e = DummyDescriptorElement(expected_type, expected_uid)
        state = e.__getstate__()
        self.assertDictEqual(
            state,
            {
                '_type_label': expected_type,
                '_uuid': expected_uid,
            }
        )

    def test_setstate(self):
        # Intentionally bad input types.
        # noinspection PyTypeChecker
        e = DummyDescriptorElement(None, None)
        self.assertIsNone(e._type_label)
        self.assertIsNone(e._uuid)

        expected_type = 'a'
        expected_uid = 'b'
        state = {
            '_type_label': expected_type,
            '_uuid': expected_uid,
        }
        e.__setstate__(state)
        self.assertEqual(e._type_label, expected_type)
        self.assertEqual(e._uuid, expected_uid)



import abc

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.utils import (
    bin_utils,
    parallel,
)

from ._defaults import DFLT_CLASSIFIER_FACTORY


class Classifier (SmqtkAlgorithm):
    """
    Interface for algorithms that classify input descriptors into discrete
    labels and/or label confidences.
    """

    def classify(self, d, factory=DFLT_CLASSIFIER_FACTORY, overwrite=False):
        """
        Classify the input descriptor against one or more discrete labels,
        outputting a ClassificationElement containing the classification
        result.

        We return confidence values for each label the configured model
        contains. Implementations may act in a discrete manner whereby only one
        label is marked with a ``1`` value (others being ``0``), or in a
        continuous manner whereby each label is given a confidence-like value
        in the [0, 1] range.

        The returned ``ClassificationElement`` will have the same UUID as the
        input ``DescriptorElement``.

        :param d: Input descriptor to classify
        :type d: smqtk.representation.DescriptorElement

        :param factory: Classification element factory. The default factory
            yields MemoryClassificationElement instances.
        :type factory: smqtk.representation.ClassificationElementFactory

        :param overwrite: Recompute classification of the input descriptor and
            set the results to the ClassificationElement produced by the
            factory.
        :type overwrite: bool

        :raises ValueError: The given descriptor element did not have a vector
            to operate on.
        :raises RuntimeError: Could not perform classification for some reason
            (see message in raised exception).

        :return: Classification result element
        :rtype: smqtk.representation.ClassificationElement

        """
        if not d.has_vector():
            raise ValueError("Given DescriptorElement does not contain a "
                             "vector to classify.")
        c_elem = factory.new_classification(self.name, d.uuid())
        if overwrite or not c_elem.has_classifications():
            c = self._classify(d)
            c_elem.set_classification(c)

        return c_elem

    def classify_async(self, d_iter, factory=DFLT_CLASSIFIER_FACTORY,
                       overwrite=False, procs=None, use_multiprocessing=False,
                       ri=None):
        """
        Asynchronously classify the DescriptorElements in the given iterable.

        :param d_iter: Iterable of DescriptorElements
        :type d_iter:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param factory: Classifier element factory to use for element
            generation. The default factory yields MemoryClassificationElement
            instances.
        :type factory: smqtk.representation.ClassificationElementFactory

        :param overwrite: Recompute classification of the input descriptor and
            set the results to the ClassificationElement produced by the
            factory.
        :type overwrite: bool

        :param procs: Explicit number of cores/thread/processes to use.
        :type procs: None | int

        :param use_multiprocessing: Use multiprocessing instead of threading.
        :type use_multiprocessing: bool

        :param ri: Progress reporting interval in seconds. Set to a value > 0 to
            enable. Disabled by default.
        :type ri: float | None

        :return: Mapping of input DescriptorElement instances to the computed
            ClassificationElement. ClassificationElement UUID's are congruent
            with the UUID of the DescriptorElement
        :rtype: dict[smqtk.representation.DescriptorElement,
                     smqtk.representation.ClassificationElement]

        """
        self._log.debug("Async classifying descriptors")
        ri = ri and ri > 0 and ri

        def work(d_elem):
            return d_elem, self.classify(d_elem, factory, overwrite)

        classifications = parallel.parallel_map(
            work, d_iter,
            cores=procs,
            ordered=False,
            use_multiprocessing=use_multiprocessing,
        )

        r_state = [0] * 7
        if ri:
            r_progress = bin_utils.report_progress
        else:
            def r_progress(*_):
                return

        d2c_map = {}
        for d, c in classifications:
            d2c_map[d] = c

            r_progress(self._log.debug, r_state, ri)

        return d2c_map

    #
    # Abstract methods
    #

    @abc.abstractmethod
    def get_labels(self):
        """
        Get the sequence of class labels that this classifier can classify
        descriptors into. This includes the negative label.

        :return: Sequence of possible classifier labels.
        :rtype: collections.Sequence[collections.Hashable]

        :raises RuntimeError: No model loaded.

        """

    @abc.abstractmethod
    def _classify(self, d):
        """
        Internal method that constructs the label-to-confidence map (dict) for
        a given DescriptorElement.

        The passed descriptor element is guaranteed to have a vector to extract.
        It is not extracted yet due to the philosophy of waiting until the
        vector is immediately needed. This moment is thus determined by the
        implementing algorithm.

        :param d: DescriptorElement containing the vector to classify.
        :type d: smqtk.representation.DescriptorElement

        :raises RuntimeError: Could not perform classification for some reason
            (see message in raised exception).

        :return: Dictionary mapping trained labels to classification confidence
            values
        :rtype: dict[collections.Hashable, float]

        """



"""
SMQTK Web Applications
"""

import inspect
import os

import flask

import smqtk.utils
from smqtk.utils import SmqtkObject
from smqtk.utils import plugin
from smqtk.utils import merge_dict


# noinspection PyAbstractClass
class SmqtkWebApp (SmqtkObject, flask.Flask,
                   smqtk.utils.Configurable, plugin.Pluggable):
    """
    Base class for SMQTK web applications
    """

    @classmethod
    def impl_directory(cls):
        """
        :return: Directory in which this implementation is contained.
        :rtype: str
        """
        return os.path.dirname(os.path.abspath(inspect.getfile(cls)))

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        This should be overridden in each implemented application class to add
        appropriate configuration.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        return {
            "flask_app": {
                "SECRET_KEY": "MySuperUltraSecret",
                "BASIC_AUTH_USERNAME": "demo",
                "BASIC_AUTH_PASSWORD": "demo"
            },
            "server": {
                'host': "127.0.0.1",
                'port': 5000
            }
        }

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Override to just pass the configuration dictionary to constructor
        """
        # Repeated from super method due to overriding how constructor is called
        if merge_default:
            merged = cls.get_default_config()
            merge_dict(merged, config_dict)
            config_dict = merged
        return cls(config_dict)

    def __init__(self, json_config):
        """
        Initialize application based of supplied JSON configuration

        :param json_config: JSON configuration dictionary
        :type json_config: dict

        """
        super(SmqtkWebApp, self).__init__(
            self.__class__.__name__,
            static_folder=os.path.join(self.impl_directory(), 'static'),
            template_folder=os.path.join(self.impl_directory(), 'templates')
        )

        #
        # Configuration setup
        #
        self.json_config = json_config

        # Factor 'flask_app' configuration properties into self.config
        for k in self.json_config['flask_app']:
            self.config[k] = self.json_config['flask_app'][k]

        #
        # Security
        #
        self.secret_key = self.config['SECRET_KEY']

    def get_config(self):
        return self.json_config

    def run(self, host=None, port=None, debug=False, **options):
        """
        Override of the run method, drawing running host and port from
        configuration by default. 'host' and 'port' values specified as argument
        or keyword will override the app configuration.
        """
        super(SmqtkWebApp, self)\
            .run(host=(host or self.json_config['server']['host']),
                 port=(port or self.json_config['server']['port']),
                 debug=debug,
                 **options)


def get_web_applications(reload_modules=False):
    """
    Discover and return SmqtkWebApp implementation classes found in the plugin
    directory. Keys in the returned map are the names of the discovered classes
    and the paired values are the actual class type objects.

    We look for modules (directories or files) that start with and alphanumeric
    character ('_' prefixed files/directories are hidden, but not recommended).

    Within a module, we first look for a helper variable by the name
    ``APPLICATION_CLASS``, which can either be a single class object or
    an iterable of class objects, to be exported. If the variable is set to
    None, we skip that module and do not import anything. If the variable is not
    present, we look for a class by the same na e and casing as the module's
    name. If neither are found, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class objects of type ``SmqtkWebApp`` whose
        keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    import os
    from smqtk.utils.plugin import get_plugins

    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "APPLICATION_PATH"
    helper_var = "APPLICATION_CLASS"
    return get_plugins(__name__, this_dir, env_var, helper_var, SmqtkWebApp,
                       reload_modules=reload_modules)



from __future__ import (absolute_import, division, print_function)

import numpy as np


def adjust_proba(proba, adj):
    """
    Adjust the probabilities in `proba` with the log-scale adjustments `adj`.
    If the probabilities in a row of `proba` are `(p1,...,pd)`, and the
    adjustments are `(a1,...,ad)`, then the output will be:

        `(p1*exp(a1),...,pd*exp(ad)) / sum(p1*exp(a1),...,pd*exp(ad))`.

    The output is therefore scale-invariant in `proba`, so `proba` need not
    sum to one, but must contain all nonnegative values and at least one
    element per row must be positive.

    This can be used to modify predictions from (e.g.) classifiers, such as
    when the training sizes are imbalanced. Applying
    `(-log(card1),...,-log(cardd)`, where `cardi` is the cardinality of the
    class in the corresponding training set, would lower the recall for high
    cardinality classes and boost the recall for low cardinality classes.[1]

    An alternative interpretation of this function is that if a row of `proba`
    is a "prior," and `adj` is the logarithm of the "evidence," then the
    corresponding output row is the "posterior" after applying a Bayesian
    update. In the example above, our evidence is that the classes are skewed
    so we want to adjust the output to compensate.

    The results are *additive in `adj`* for multiple calls of the function.
    That is, `adjust_proba(adjust_proba(proba, adj1), adj2)` is simply
    `adjust_proba(proba, adj1+adj2)`. This means that subsequent calls are
    unnecessary and adjustments may be accumulated and applied all at once.
    (This follows from the fact that successive Bayesian updates are
    equivalent to updating against the product of the evidence.)

    It is also the case that if `adj` is a constant vector, then
    `np.allclose(adjust_proba(proba, adj), proba)` will return true. (This
    follows from the fact that uniform evidence has no effect in Bayesian
    updates.) Combined with the previous fact, this means that `adj` may take
    values of either (1,0,0) or (0,-1,-1), for example, and produce the same
    output for the same `proba` (since `(1,0,0) = (1,1,1) + (0,-1,-1)`).

    :param proba: Numpy array of shape (n,d) with probabilities in each row.
       Each column represents a class. The probabilities need not be
       normalized to sum to 1.
    :type proba: np.ndarray | collections.Iterable

    :param adj: An iterable with d elements or a numpy array of shape (d,)
        that contains the adjustments for each class. The adjustments are
        log-scale and the probabilities in `proba` will be multiplied by the
        exponential of `adj`, then normalized. The parameters in `adj` need
        not be normalized.
    :type adj: np.ndarray | collections.Iterable

    :return: A numpy array with the same shape as `proba` containing the
        adjusted probabilities per-class. The output will sum to one on the
        second axis.
    :rtype: np.ndarray

    [1] Note: If `proba` is a set of probabilistic predictions of a calibrated
    classifier, it is likely that applying adjustments to the predictions will
    affect accuracy negatively.
    """
    adj_proba, adj_vals = np.atleast_2d(proba, adj)

    if adj_proba.shape[1] != adj_vals.shape[1]:
        raise ValueError(
            "The dimensions of probabilities and adjustments must be "
            "compatible (number of columns must be equal).")

    if np.any(adj_proba < 0):
        raise ValueError("Probabilities must be at least 0.")
    # After eliminating the possibility of a negative value, we only need to
    # test if the sum is close to zero
    if np.isclose(adj_proba.sum(), 0):
        raise ValueError("At least one probability must be positive.")

    # Prevent exponential overflow (this cancels out on return)
    adj_proba = adj_proba * np.exp(adj_vals - adj_vals.max())

    # Normalize and return adjusted probabilities
    return adj_proba / adj_proba.sum(axis=1, keepdims=True)



import mock
import unittest

from smqtk.utils.file_utils import safe_file_write


class TestSafeFileWrite (unittest.TestCase):
    """
    Tests for the ``smqtk.utils.file_utils.safe_file_write`` function.

    Mocking out underlying function that would have filesystem side effects.
    """

    @mock.patch('smqtk.utils.file_utils.safe_create_dir')
    @mock.patch('smqtk.utils.file_utils.os.rename')
    @mock.patch('smqtk.utils.file_utils.os.remove')
    @mock.patch('smqtk.utils.file_utils.tempfile.NamedTemporaryFile')
    def test_safe_file_write_relative_simple(
            self, m_NTF, m_remove, m_rename, m_scd):
        # Experimental filepath and content.
        fp = 'bar.txt'
        expected_bytes = 'hello world'

        # Mock return for temp file creation so we can check os.* calls.
        m_file = m_NTF.return_value
        test_tmp_fp = 'temp fp'
        m_file.name = test_tmp_fp

        safe_file_write(fp, expected_bytes)

        m_scd.assert_called_once_with('')
        m_NTF.assert_called_once_with(suffix='.txt', prefix='bar.', dir='', delete=False)
        m_file.write.assert_called_once_with(expected_bytes)
        m_file.__exit__.assert_called_once_with(None, None, None)
        self.assertEqual(m_remove.call_count, 0)
        m_rename.assert_called_once_with(test_tmp_fp, fp)

    @mock.patch('smqtk.utils.file_utils.safe_create_dir')
    @mock.patch('smqtk.utils.file_utils.os.rename')
    @mock.patch('smqtk.utils.file_utils.os.remove')
    @mock.patch('smqtk.utils.file_utils.tempfile.NamedTemporaryFile')
    def test_safe_file_write_relative_subdir(
            self, m_NTF, m_remove, m_rename, m_scd):
        # Experimental filepath and content.
        fp = 'foo/other/bar.txt'
        expected_bytes = 'hello world'

        # Mock return for temp file creation so we can check os.* calls.
        m_file = m_NTF.return_value
        test_tmp_fp = 'temp fp'
        m_file.name = test_tmp_fp

        safe_file_write(fp, expected_bytes)

        m_scd.assert_called_once_with('foo/other')
        m_NTF.assert_called_once_with(suffix='.txt', prefix='bar.',
                                      dir='foo/other', delete=False)
        m_file.write.assert_called_once_with(expected_bytes)
        m_file.__exit__.assert_called_once_with(None, None, None)
        self.assertEqual(m_remove.call_count, 0)
        m_rename.assert_called_once_with(test_tmp_fp, fp)

    @mock.patch('smqtk.utils.file_utils.safe_create_dir')
    @mock.patch('smqtk.utils.file_utils.os.rename')
    @mock.patch('smqtk.utils.file_utils.os.remove')
    @mock.patch('smqtk.utils.file_utils.tempfile.NamedTemporaryFile')
    def test_safe_file_write_custom_tmp_dir(
            self, m_NTF, m_remove, m_rename, m_scd):
        # Experimental filepath and content.
        fp = 'foo/other/bar.txt'
        expected_bytes = 'hello world'
        custom_tmp_dir = '/some/other/directory'

        # Mock return for temp file creation so we can check os.* calls.
        m_file = m_NTF.return_value
        test_tmp_fp = 'temp fp'
        m_file.name = test_tmp_fp

        safe_file_write(fp, expected_bytes, custom_tmp_dir)

        m_scd.assert_called_once_with('foo/other')
        m_NTF.assert_called_once_with(suffix='.txt', prefix='bar.',
                                      dir=custom_tmp_dir, delete=False)
        m_file.write.assert_called_once_with(expected_bytes)
        m_file.__exit__.assert_called_once_with(None, None, None)
        self.assertEqual(m_remove.call_count, 0)
        m_rename.assert_called_once_with(test_tmp_fp, fp)

    @mock.patch('smqtk.utils.file_utils.safe_create_dir')
    @mock.patch('smqtk.utils.file_utils.os.rename')
    @mock.patch('smqtk.utils.file_utils.os.remove')
    @mock.patch('smqtk.utils.file_utils.tempfile.NamedTemporaryFile')
    def test_safe_file_write_absolute(
            self, m_NTF, m_remove, m_rename, m_scd):
        # Experimental filepath and content.
        fp = '/some/absolute/dir/bar.txt'
        expected_bytes = 'hello world'

        # Mock return for temp file creation so we can check os.* calls.
        m_file = m_NTF.return_value
        test_tmp_fp = 'temp fp'
        m_file.name = test_tmp_fp

        safe_file_write(fp, expected_bytes)

        m_scd.assert_called_once_with('/some/absolute/dir')
        m_NTF.assert_called_once_with(suffix='.txt', prefix='bar.',
                                      dir='/some/absolute/dir', delete=False)
        m_file.write.assert_called_once_with(expected_bytes)
        m_file.__exit__.assert_called_once_with(None, None, None)
        self.assertEqual(m_remove.call_count, 0)
        m_rename.assert_called_once_with(test_tmp_fp, fp)

    @mock.patch('smqtk.utils.file_utils.safe_create_dir')
    @mock.patch('smqtk.utils.file_utils.os.rename')
    @mock.patch('smqtk.utils.file_utils.os.remove')
    @mock.patch('smqtk.utils.file_utils.tempfile.NamedTemporaryFile')
    def test_safe_file_write_raising_write(
            self, m_NTF, m_remove, m_rename, m_scd):
        # Test for what happens when file.write raises an exception.

        # Experimental filepath and content.
        fp = 'bar.txt'
        expected_bytes = 'hello world'

        # Mock return for temp file creation so we can check os.* calls.
        m_file = m_NTF.return_value
        test_tmp_fp = 'temp fp'
        m_file.name = test_tmp_fp
        # Mock return from write simulating not all bytes being written.
        m_file.write.side_effect = OSError

        self.assertRaises(
            OSError,
            safe_file_write, fp, expected_bytes
        )

        m_scd.assert_called_once_with('')
        m_NTF.assert_called_once_with(suffix='.txt', prefix='bar.', dir='', delete=False)
        m_file.write.assert_called_once_with(expected_bytes)
        # Remove should now be called on temp file path
        self.assertEqual(m_remove.call_count, 1)
        m_remove.assert_called_once_with(test_tmp_fp)
        self.assertEqual(m_file.__exit__.call_count, 1)
        # Rename should no longer be called.
        self.assertEqual(m_rename.call_count, 0)



import six
from six.moves.urllib_parse import urlparse

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation import DataElement

try:
    import girder_client
except ImportError:
    girder_client = None


__all__ = [
    'GirderDataElement',
]


class GirderDataElement (DataElement):
    """
    Element whose data is stored via a Girder backend.  Accesses via Girder
    REST API given user credentials.
    """

    @classmethod
    def is_usable(cls):
        """
        :return: If this element type is usable
        :rtype: bool
        """
        return girder_client is not None

    # TODO: from_uri
    #       - maybe optionally allow API key in place of user/pass spec
    #           (i.e. girder://<api_key>@<url...>
    #       - <url> in above I guess would be the api/v1/... URL, including any
    #           parameters needed
    @classmethod
    def from_uri(cls, uri):
        """
        Creates a ``GirderDataElement`` from ``uri``.

        The parsing is accomplished by passing ``uri`` to ``urlparse``.
        This allows for quite a bit of flexibility in the types of URIs that
        can be passed. Valid Girder URIs are:

        (a) girder://token:<token>@<api_url>/file/<file_id>
        (b) girder://api_key:<api_key>@<api_url>/file/<file_id>

        Currently, only GirderDataElements can be built from Files, so the
        URL should end in /file/{id}.

        :return: Data element created from ``uri``
        :rtype: GirderDataElement

        :raises ValueError: An invalid URI was passed.

        :raises AssertionError: If the path parsed from a URI of the form (a)
            does not have '/file/' as its penultimate location before the
            identifier.

        """
        # urlparse seems to not be phased by the 'girder' protocol instead of
        # http, so no replacing needs to be done.
        parsed_uri = urlparse(uri)
        if parsed_uri.scheme != 'girder':
            raise InvalidUriError(uri, 'Invalid Girder URI. Girder URIs must '
                                       'start with girder://')

        if not parsed_uri.netloc:
            raise InvalidUriError(uri, 'No parsed netloc from given URI.')

        token = api_key = None

        if '@' in parsed_uri.netloc:
            credentials, scheme = parsed_uri.netloc.split('@')
            cred_type, cred = credentials.split(':')

            if cred_type == 'token':
                token = cred
            elif cred_type == 'api_key':
                api_key = cred
        else:
            scheme = parsed_uri.netloc

        try:
            path, file_id = parsed_uri.path.split('/file/')
        except ValueError:
            raise InvalidUriError(uri, 'Invalid Girder URI. Girder URIs must '
                                       'contain a /file/<file_id> segment.')

        return cls(file_id, '%s%s' % (scheme, path), api_key, token)

    # NOTE: This usage of api "root" contradicts girder client's notion of the
    #       api root
    def __init__(self, file_id, api_root='http://localhost:8080/api/v1',
                 api_key=None, token=None):
        """
        Initialize data element to point to a specific file hosted in Girder

        An authorization token will be generated if an API key is provided at
        construction.  A new token will be requested if it has expired.

        :param file_id: ID of the file in Girder
        :type file_id: str

        :param api_root: Girder API root URL
        :type api_root: str

        :param api_key: Optional API key to request token with. Otherwise an
        :type api_key:

        """
        # TODO: Template sub-URLs for customizing model/download endpoints used?
        #       e.g. 'file/{file_id:s}' and '
        super(GirderDataElement, self).__init__()

        self.file_id = file_id
        self.api_root = api_root
        self.token = token
        self.api_key = api_key

        # compat with DataFileElement (compute_many_descriptors needs this)
        self._filepath = self.file_id

        self.gc = girder_client.GirderClient(apiUrl=api_root)

        if token is not None:
            self.gc.token = token
        elif api_key is not None:
            self.gc.authenticate(apiKey=api_key)

        # Cache so we don't have to query server multiple times for multiple
        # calls.
        self._content_type = None

    def __repr__(self):
        return super(GirderDataElement, self).__repr__() + \
            "{file_id: %s, api_root: %s, api_key: %s, token: %s}" % (
                self.file_id, self.api_root, self.api_key or '',
                self.token or ''
            )

    def get_config(self):
        return {
            'file_id': self.file_id,
            'api_root': self.api_root,
            'api_key': self.api_key,
            'token': self.token
        }

    def content_type(self):
        if self._content_type is None:
            self._log.debug("Getting content type for file ID %s"
                            % self.file_id)
            file_model = self.get_file_model()

            if file_model is not None:
                self._content_type = file_model['mimeType']

        return self._content_type

    def get_file_model(self):
        """
        Get the file model json from the server.

        Returns None if the file can't be retrieved from the server.

        :return: file model model as a dictionary
        :rtype: dict | None

        """
        try:
            return self.gc.getFile(self.file_id)
        except girder_client.HttpError:
            return None

    def is_empty(self):
        """
        Check if this element contains no bytes.

        This plugin checks that we can get a file model return from girder and
        that the size of the file queried is non-zero.

        :return: If there is a model for our item or if our item contains 0
            bytes.
        :rtype: bool

        """
        m = self.get_file_model()
        return m is None or m['size'] == 0

    def get_bytes(self):
        """
        Get the bytes of the file stored in girder.

        :return: Get the byte stream for this data element.
        :rtype: bytes
        """
        # Download file bytes from girder
        content = six.BytesIO()
        self._log.debug("Getting bytes for file ID %s", self.file_id)
        self.gc.downloadFile(self.file_id, content)
        return bytes(content.getvalue())

    def writable(self):
        """
        Determine if a Girder file is able to be written to. Note that this
        requires inferring the access level by traversing to the parent folder
        since this is how Girder determines access.

        :return: if this instance supports setting bytes.
        :rtype: bool
        """
        file_model = self.get_file_model()

        if file_model is None:
            return False
        else:
            item_model = self.gc.getItem(file_model['itemId'])
            folder_model = self.gc.getFolder(item_model['folderId'])

            # See girder.constants.AccessType
            return folder_model['_accessLevel'] >= 1

    def set_bytes(self, b):
        """
        Set bytes to this data element in the form of a string.

        Not all implementations may support setting bytes (writing). See the
        ``writable`` method.

        :param b: bytes to set.
        :type b: str

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        if not self.writable():
            raise ReadOnlyError('Unauthorized access to write to Girder file %s'
                                % self.file_id)

        try:
            # noinspection PyTypeChecker
            self.gc.uploadFileContents(self.file_id, six.BytesIO(b), len(b))
        except girder_client.HttpError as e:
            if e.status == 401:
                raise ReadOnlyError('Unauthorized access to write to Girder '
                                    'file %s' % self.file_id)
            else:
                raise e



"""
Utilities for URLs
"""
import re


PROTOCOL_HEADER = re.compile('\w+://')
URL_SEP = '/'


def url_join(url, *urls):
    """
    Join one or more URL components intelligently.

    The return is the concatenation of ``url`` and any members of ``*urls`` with
    exactly one slash ("/") following each non-empty part except the last,
    meaning the result will only end in a trailing slash if the last part is
    empty. If a component starts with a protocol-like string or a leading slash,
    all previous components are thrown away and joining continues from that
    part.

    We define "protocol-like" as something matching the regex "\w+://".

    Leading slashes reduced to a single leading slash if multiple are present.

    This function is different in that we allow the joining of relative URLs
    that do not start with a protocol component.

    :param url: First URL component
    :type url: str

    :param urls: path components to join with slashes.
    :type urls: string-castable

    :return: Joined URL string
    :rtype: str

    """
    # State-builder methods
    urls = (url,) + urls
    concat = ''
    last_empty = False
    first = True
    skip_slash_prefix = True
    for p in urls:
        p = str(p)
        if not p:
            last_empty = True
        else:
            last_empty = False

            # Check protocol header presence
            m = PROTOCOL_HEADER.search(p)
            if m and m.start() == 0:
                protocol = p[m.start():m.end()]
                tail = p[m.end():].rstrip(URL_SEP)
                # Don't add separator on next if just protocol header
                skip_slash_prefix = not tail

                concat = protocol + tail
                # print("Restarting:", concat)

            # Check for leading slash
            elif p.startswith(URL_SEP):
                p = p.strip(URL_SEP)
                concat = URL_SEP + p
                # Don't add separator on next if p was just a slash
                skip_slash_prefix = not p
                # print("Restarting:", concat)

            # Otherwise append
            else:
                p = p.rstrip(URL_SEP)
                if first or skip_slash_prefix:
                    concat += p
                    skip_slash_prefix = False
                else:
                    concat += URL_SEP + p
                # print("Appended:", p,  "->", concat)

            first = False

    if concat and last_empty:
        concat += URL_SEP

    return concat



import mock
import os
import sys
import tempfile
import unittest

from six.moves import StringIO

from smqtk.bin.check_images import main as check_images_main
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.tests import TEST_DATA_DIR
from smqtk.utils.image_utils import is_loadable_image, is_valid_element


class TestIsLoadableImage(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        cls.good_image = DataFileElement(os.path.join(TEST_DATA_DIR,
                                                      'Lenna.png'))
        cls.non_image = DataFileElement(os.path.join(TEST_DATA_DIR,
                                                     'test_file.dat'))

    def test_non_data_element_raises_exception(self):
        # should throw:
        # AttributeError: 'bool' object has no attribute 'get_bytes'
        self.assertRaises(
            AttributeError,
            is_loadable_image, False
        )

    def test_unloadable_image_returns_false(self):
        assert is_loadable_image(self.non_image) is False

    def test_loadable_image_returns_true(self):
        assert is_loadable_image(self.good_image) is True


class TestIsValidElement(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        cls.good_image = DataFileElement(os.path.join(TEST_DATA_DIR,
                                                      'Lenna.png'))
        cls.non_image = DataFileElement(os.path.join(TEST_DATA_DIR,
                                                     'test_file.dat'))

    def test_non_data_element(self):
        # Should check that input datum is a DataElement instance.
        # noinspection PyTypeChecker
        assert is_valid_element(False) is False

    def test_invalid_content_type(self):
        assert is_valid_element(self.good_image, valid_content_types=[]) \
               is False

    def test_valid_content_type(self):
        assert is_valid_element(self.good_image,
                                valid_content_types=['image/png']) is True

    def test_invalid_image_returns_false(self):
        assert is_valid_element(self.non_image, check_image=True) is False


class TestCheckImageCli(unittest.TestCase):

    @staticmethod
    def check_images():
        """ Simulate execution of check_images utility main. """
        saved_stdout, saved_stderr = sys.stdout, sys.stderr

        out, err = StringIO(), StringIO()
        try:
            sys.stdout, sys.stderr = out, err
            check_images_main()
        except SystemExit as ex:
            print("Encountered SystemExit exception, code {}".format(ex.code))
        finally:
            stdout, stderr = out.getvalue().strip(), err.getvalue().strip()
            sys.stdout, sys.stderr = saved_stdout, saved_stderr

        return stdout, stderr

    def test_base_case(self):
        # noinspection PyUnresolvedReferences
        with mock.patch.object(sys, 'argv', ['']):
            assert 'Validate a list of images returning the filepaths' in \
                self.check_images()[0]

    def test_check_images(self):
        # Create test file with a valid, invalid, and non-existent image
        _, filename = tempfile.mkstemp()

        with open(filename, 'w') as outfile:
            outfile.write(os.path.join(TEST_DATA_DIR, 'Lenna.png') + '\n')
            outfile.write(os.path.join(TEST_DATA_DIR, 'test_file.dat') + '\n')
            outfile.write(os.path.join(TEST_DATA_DIR, 'non-existent-file.jpeg'))

        # noinspection PyUnresolvedReferences
        with mock.patch.object(sys, 'argv', ['', '--file-list', filename]):
            out, err = self.check_images()

            assert out == ','.join([os.path.join(TEST_DATA_DIR, 'Lenna.png'),
                                    '3ee0d360dc12003c0d43e3579295b52b64906e85'])
            assert 'non-existent-file.jpeg' not in out

        # noinspection PyUnresolvedReferences
        with mock.patch.object(sys, 'argv',
                               ['', '--file-list', filename, '--invert']):
            out, err = self.check_images()

            assert out == ','.join([os.path.join(TEST_DATA_DIR,
                                                 'test_file.dat'),
                                    'da39a3ee5e6b4b0d3255bfef95601890afd80709'])
            assert 'non-existent-file.jpeg' not in out



import threading

import six
from six.moves import cPickle as pickle

from smqtk.representation.data_element import get_data_element_impls
from smqtk.representation.key_value import KeyValueStore, NO_DEFAULT_VALUE
from smqtk.utils.plugin import make_config, from_plugin_config, to_plugin_config


class MemoryKeyValueStore (KeyValueStore):
    """
    Thread-safe in-memory implementation of KeyValueStore interface with
    optional caching from/to a ``DataElement`` instance.

    Any keys and values compatible with a standard python dictionary are
    compatible with this implementation.

    **WARNING:** *This element uses pickle serialization for storing cached
    bytes. This is a security risk when not using secured or authorized storage
    mediums.*

    """

    PICKLE_PROTOCOL = 2

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        default = super(MemoryKeyValueStore, cls).get_default_config()
        default['cache_element'] = make_config(get_data_element_impls())
        return default

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: MemoryKeyValueStore

        """
        # Copy top-level of config in order to not modify input instance.
        c = config_dict.copy()
        # Simplify specification for "no cache element"
        if 'cache_element' not in c or \
                c['cache_element'] is None or \
                c['cache_element']['type'] is None:
            c['cache_element'] = None
        else:
            # Create from nested config.
            c['cache_element'] = \
                from_plugin_config(config_dict['cache_element'],
                                   get_data_element_impls())
        return super(MemoryKeyValueStore, cls).from_config(c)

    def __init__(self, cache_element=None):
        """
        Create new in-memory key-value store with optional cache data element.

        This element is read-only when using a DataElement instance for a cache
        and that element is not writable.

        :param cache_element: Optional data element to load/save state from/to.
        :type cache_element: smqtk.representation.DataElement

        """
        super(MemoryKeyValueStore, self).__init__()
        self._cache_element = cache_element
        self._table = {}
        self._table_lock = threading.RLock()

        # Only try to load from cache if the cache has any bytes to try to
        # deserialize.
        if self._cache_element:
            c_bytes = self._cache_element.get_bytes()
            if c_bytes:
                self._table = pickle.loads(c_bytes)

    def __repr__(self):
        return super(MemoryKeyValueStore, self).__repr__() \
            % ("cache_element: %s" % repr(self._cache_element))

    def cache_table(self):
        """
        Cache the current table to the currently set cache element.

        If there is no cache element, this method does nothing.
        """
        if self._cache_element is not None:
            # TODO(paul.tunison): Some other serialization than Pickle.
            #   - pickle loading allows arbitrary code execution on host.
            self._log.debug("Caching table to {}".format(self._cache_element))
            self._cache_element.set_bytes(
                pickle.dumps(self._table, self.PICKLE_PROTOCOL))

    def count(self):
        with self._table_lock:
            return len(self._table)

    def get_config(self):
        # Recursively get config from data element if we have one.
        if hasattr(self._cache_element, 'get_config'):
            elem_config = to_plugin_config(self._cache_element)
        else:
            # No cache element, output default config with no type.
            elem_config = make_config(get_data_element_impls())
        return {
            'cache_element': elem_config
        }

    def keys(self):
        """
        :return: Iterator over keys in this store.
        :rtype: __generator[collections.Hashable]
        """
        return six.iterkeys(self._table)

    def is_read_only(self):
        """
        If this element is read-only or not as determined by any cache element's
        read-only status.

        :return: True if this instance is using a ``DataElement``
        :rtype: bool
        """
        # Only not writable if we have a cache and it is not writable
        if self._cache_element and not self._cache_element.writable():
            return True
        return False

    def has(self, key):
        """
        Check if this store has a value for the given key.

        :param key: Key to check for a value for.
        :type key: collections.Hashable

        :return: If this store has a value for the given key.
        :rtype: bool

        """
        return key in self._table

    def add(self, key, value):
        """
        Add a key-value pair to this store.

        :param key: Key for the value. Must be hashable.
        :type key: collections.Hashable

        :param value: Python object to store.
        :type value: object

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        super(MemoryKeyValueStore, self).add(key, value)
        with self._table_lock:
            self._table[key] = value
            self.cache_table()
        return self

    def add_many(self, d):
        """
        Add multiple key-value pairs at a time into this store as represented in
        the provided dictionary `d`.

        :param d: Dictionary of key-value pairs to add to this store.
        :type d: dict[collections.Hashable, object]

        :return: Self.
        :rtype: MemoryKeyValueStore

        """
        super(MemoryKeyValueStore, self).add_many(d)
        with self._table_lock:
            for k, v in six.iteritems(d):
                self._table[k] = v
            self.cache_table()
        return self

    def remove(self, key):
        """
        Remove a single key-value entry.

        :param key: Key to remove.
        :type key: collections.Hashable

        :raises KeyError: The given key is not present in this store and no
            default value given.

        :return: Self.
        :rtype: KeyValueStore

        """
        super(MemoryKeyValueStore, self).remove(key)
        with self._table_lock:
            del self._table[key]
            self.cache_table()
        return self

    def remove_many(self, keys):
        """
        Remove multiple keys and associated values.

        :param keys: Iterable of keys to remove.  If this is empty this method
            does nothing.
        :type keys: collections.Iterable[collections.Hashable]

        :raises KeyError: The given key is not present in this store and no
            default value given.  The store is not modified if any key is
            invalid.

        :return: Self.
        :rtype: KeyValueStore

        """
        super(MemoryKeyValueStore, self).remove_many(keys)
        with self._table_lock:
            # Make sure all keys are represented
            key_diff = set(keys) - set(self._table)
            if key_diff:
                if len(key_diff) == 1:
                    raise KeyError(list(key_diff)[0])
                else:
                    raise KeyError(key_diff)
            # Actually remove keys.
            for k in keys:
                del self._table[k]
            self.cache_table()
        return self

    def get(self, key, default=NO_DEFAULT_VALUE):
        """
        Get the value for the given key.

        :param key: Key to get the value of.
        :type key: collections.Hashable

        :param default: Optional default value if the given key is not present
            in this store. This may be any value except for the
            ``NO_DEFAULT_VALUE`` constant (custom anonymous class instance).
        :type default: object

        :raises KeyError: The given key is not present in this store and no
            default value given.

        :return: Deserialized python object stored for the given key.
        :rtype: object

        """
        with self._table_lock:
            if default is NO_DEFAULT_VALUE:
                return self._table[key]
            else:
                return self._table.get(key, default)

    def clear(self):
        """
        Clear this key-value store.

        :return: Self.
        :rtype: KeyValueStore

        """
        super(MemoryKeyValueStore, self).clear()
        with self._table_lock:
            self._table.clear()
        return self



import mock
import unittest

import six

from smqtk.algorithms.nn_index.hash_index import HashIndex, get_hash_index_impls


class DummyHI (HashIndex):

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        """ stub """

    def count(self):
        return 0

    def _build_index(self, hashes):
        """ stub """

    def _update_index(self, hashes):
        """ stub """

    def _remove_from_index(self, hashes):
        """ stub """

    def _nn(self, h, n=1):
        """ stub """


class TestHashIndex (unittest.TestCase):

    def test_get_impls(self):
        m = get_hash_index_impls()
        self.assertIsInstance(m, dict)
        for cls in six.itervalues(m):
            self.assertTrue(issubclass(cls, HashIndex))

    def test_empty_iterable_exception(self):
        v = DummyHI._empty_iterable_exception()
        self.assertIsInstance(v, ValueError)
        self.assertRegexpMatches(str(v), "hash vectors")

    def test_build_index_empty_iter(self):
        idx = DummyHI()
        idx._build_index = mock.MagicMock()
        self.assertRaisesRegexp(
            ValueError,
            str(HashIndex._empty_iterable_exception()),
            idx.build_index, []
        )
        # Internal method should not have been called
        idx._build_index.assert_not_called()

    def test_build_index_with_values(self):
        idx = DummyHI()
        idx._build_index = mock.MagicMock()
        # No error should be returned. Returned iterable contents should match
        # input values.
        # noinspection PyTypeChecker
        idx.build_index([0, 1, 2])
        self.assertSetEqual(
            set(idx._build_index.call_args[0][0]),
            {0, 1, 2}
        )

    def test_update_index_empty_iter(self):
        idx = DummyHI()
        idx._update_index = mock.MagicMock()
        self.assertRaisesRegexp(
            ValueError,
            "No hash vectors.*",
            idx.update_index, []
        )
        # Internal method should not have been called.
        idx._update_index.assert_not_called()

    def test_update_index_with_values(self):
        idx = DummyHI()
        idx._update_index = mock.MagicMock()
        # No error should be returned. Returned iterable contents should match
        # input values.
        # noinspection PyTypeChecker
        idx.update_index([0, 1, 2])
        self.assertSetEqual(
            set(idx._update_index.call_args[0][0]),
            {0, 1, 2}
        )

    def test_remove_from_index_empty_iter(self):
        idx = DummyHI()
        idx._remove_from_index = mock.MagicMock()
        self.assertRaisesRegexp(
            ValueError,
            "No hash vectors.*",
            idx.update_index, []
        )
        # Internal method should not have been called.
        idx._remove_from_index.assert_not_called()

    def test_remove_from_index_with_values(self):
        idx = DummyHI()
        idx._remove_from_index = mock.MagicMock()
        # No error should be returned. Returned iterable contents should match
        # input values.
        # noinspection PyTypeChecker
        idx._remove_from_index([0, 1, 2])
        self.assertSetEqual(
            set(idx._remove_from_index.call_args[0][0]),
            {0, 1, 2}
        )

    def test_nn_no_index(self):
        idx = DummyHI()
        idx._nn = mock.MagicMock()
        self.assertRaises(
            ValueError,
            idx.nn, 'something'
        )
        # Internal method should not have been called.
        idx._nn.assert_not_called()

    def test_nn_has_count(self):
        idx = DummyHI()
        idx.count = mock.MagicMock()
        idx.count.return_value = 10
        idx._nn = mock.MagicMock()
        # This call should now pass that count returns something greater than 0.
        # noinspection PyTypeChecker
        idx.nn('dummy')
        idx._nn.assert_called_with("dummy", 1)

        # noinspection PyTypeChecker
        idx.nn('bar', 10)
        idx._nn.assert_called_with("bar", 10)



from __future__ import (absolute_import, division, print_function)

import unittest

import multiprocessing
import multiprocessing.pool
import numpy
import six
from six.moves import cPickle

from smqtk.algorithms.classifier import get_classifier_impls
from smqtk.algorithms.classifier.libsvm import LibSvmClassifier
from smqtk.representation import \
    ClassificationElementFactory, \
    DescriptorElementFactory
from smqtk.representation.classification_element.memory import \
    MemoryClassificationElement
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement


if LibSvmClassifier.is_usable():

    class TestLibSvmClassifier (unittest.TestCase):

        def test_impl_findable(self):
            self.assertIn(LibSvmClassifier.__name__,
                          get_classifier_impls())

        def test_no_save_model_pickle(self):
            # Test model preservation across pickling even without model cache
            # file paths set.
            classifier = LibSvmClassifier(
                train_params={
                    '-t': 0,  # linear kernel
                    '-b': 1,  # enable probability estimates
                    '-c': 2,  # SVM-C parameter C
                    '-q': '',  # quite mode
                },
                normalize=None,  # DO NOT normalize descriptors
            )
            self.assertTrue(classifier.svm_model is None)
            # Empty model should not trigger __LOCAL__ content in pickle
            self.assertNotIn('__LOCAL__', classifier.__getstate__())
            _ = cPickle.loads(cPickle.dumps(classifier))

            # train arbitrary model (same as ``test_simple_classification``)
            DIM = 2
            N = 1000
            POS_LABEL = 'positive'
            NEG_LABEL = 'negative'
            d_factory = DescriptorElementFactory(DescriptorMemoryElement, {})
            c_factory = ClassificationElementFactory(
                MemoryClassificationElement, {}
            )

            def make_element(iv):
                i, v = iv
                d = d_factory.new_descriptor('test', i)
                d.set_vector(v)
                return d

            # Constructing artificial descriptors
            x = numpy.random.rand(N, DIM)
            x_pos = x[x[:, 1] <= 0.45]
            x_neg = x[x[:, 1] >= 0.55]
            p = multiprocessing.pool.ThreadPool()
            d_pos = p.map(make_element, enumerate(x_pos))
            d_neg = p.map(make_element, enumerate(x_neg, start=N//2))
            p.close()
            p.join()

            # Training
            classifier.train({POS_LABEL: d_pos, NEG_LABEL: d_neg})

            # Test original classifier
            t_v = numpy.random.rand(DIM)
            t = d_factory.new_descriptor('query', 0)
            t.set_vector(t_v)
            c_expected = classifier.classify(t, c_factory)

            # Should see __LOCAL__ content in pickle state now
            p_state = classifier.__getstate__()
            self.assertIn('__LOCAL__', p_state)
            self.assertIn('__LOCAL_LABELS__', p_state)
            self.assertIn('__LOCAL_MODEL__', p_state)
            self.assertTrue(len(p_state['__LOCAL_LABELS__']) > 0)
            self.assertTrue(len(p_state['__LOCAL_MODEL__']) > 0)

            # Restored classifier should classify the same test descriptor the
            # same
            #: :type: LibSvmClassifier
            classifier2 = cPickle.loads(cPickle.dumps(classifier))
            c_post_pickle = classifier2.classify(t, c_factory)
            # There may be floating point error, so extract actual confidence
            # values and check post round
            c_pp_positive = c_post_pickle[POS_LABEL]
            c_pp_negative = c_post_pickle[NEG_LABEL]
            c_e_positive = c_expected[POS_LABEL]
            c_e_negative = c_expected[NEG_LABEL]
            self.assertAlmostEqual(c_e_positive, c_pp_positive, 5)
            self.assertAlmostEqual(c_e_negative, c_pp_negative, 5)

        def test_simple_classification(self):
            """
            simple LibSvmClassifier test - 2-class

            Test libSVM classification functionality using random constructed
            data, training the y=0.5 split
            """
            DIM = 2
            N = 1000
            POS_LABEL = 'positive'
            NEG_LABEL = 'negative'
            p = multiprocessing.pool.ThreadPool()
            d_factory = DescriptorElementFactory(DescriptorMemoryElement, {})
            c_factory = ClassificationElementFactory(
                MemoryClassificationElement, {}
            )

            def make_element(iv):
                i, v = iv
                elem = d_factory.new_descriptor('test', i)
                elem.set_vector(v)
                return elem

            # Constructing artificial descriptors
            x = numpy.random.rand(N, DIM)
            x_pos = x[x[:, 1] <= 0.45]
            x_neg = x[x[:, 1] >= 0.55]

            d_pos = p.map(make_element, enumerate(x_pos))
            d_neg = p.map(make_element, enumerate(x_neg, start=N//2))

            # Create/Train test classifier
            classifier = LibSvmClassifier(
                train_params={
                    '-t': 0,  # linear kernel
                    '-b': 1,  # enable probability estimates
                    '-c': 2,  # SVM-C parameter C
                    '-q': '',  # quite mode
                },
                normalize=None,  # DO NOT normalize descriptors
            )
            classifier.train({POS_LABEL: d_pos, NEG_LABEL: d_neg})

            # Test classifier
            x = numpy.random.rand(N, DIM)
            x_pos = x[x[:, 1] <= 0.45]
            x_neg = x[x[:, 1] >= 0.55]

            d_pos = p.map(make_element, enumerate(x_pos, N))
            d_neg = p.map(make_element, enumerate(x_neg, N + N//2))

            d_pos_sync = {}  # for comparing to async
            for d in d_pos:
                c = classifier.classify(d, c_factory)
                self.assertEqual(POS_LABEL, c.max_label(),
                                 "Found False positive: %s :: %s" %
                                 (d.vector(), c.get_classification()))
                d_pos_sync[d] = c

            d_neg_sync = {}
            for d in d_neg:
                c = classifier.classify(d, c_factory)
                self.assertEqual(NEG_LABEL, c.max_label(),
                                 "Found False negative: %s :: %s" %
                                 (d.vector(), c.get_classification()))
                d_neg_sync[d] = c

            # test that async classify produces the same results
            # -- d_pos
            m_pos = classifier.classify_async(d_pos, c_factory)
            self.assertEqual(d_pos_sync, m_pos,
                             "Async computation of pos set did not yield "
                             "the same results as synchronous "
                             "classification.")
            # -- d_neg
            m_neg = classifier.classify_async(d_neg, c_factory)
            self.assertEqual(d_neg_sync, m_neg,
                             "Async computation of neg set did not yield "
                             "the same results as synchronous "
                             "classification.")
            # -- combined -- threaded
            combined_truth = dict(six.iteritems(d_pos_sync))
            combined_truth.update(d_neg_sync)
            m_combined = classifier.classify_async(
                d_pos + d_neg, c_factory,
                use_multiprocessing=False,
            )
            self.assertEqual(combined_truth, m_combined,
                             "Async computation of all test descriptors "
                             "did not yield the same results as "
                             "synchronous classification.")
            # -- combined -- multiprocess
            m_combined = classifier.classify_async(
                d_pos + d_neg, c_factory,
                use_multiprocessing=True,
            )
            self.assertEqual(combined_truth, m_combined,
                             "Async computation of all test descriptors "
                             "(mixed order) did not yield the same results "
                             "as synchronous classification.")

            # Closing resources
            p.close()
            p.join()

        def test_simple_multiclass_classification(self):
            """
            simple LibSvmClassifier test - 3-class

            Test libSVM classification functionality using random constructed
            data, training the y=0.33 and y=.66 split
            """
            DIM = 2
            N = 1000
            P1_LABEL = 'p1'
            P2_LABEL = 'p2'
            P3_LABEL = 'p3'
            p = multiprocessing.pool.ThreadPool()
            d_factory = DescriptorElementFactory(DescriptorMemoryElement, {})
            c_factory = ClassificationElementFactory(
                MemoryClassificationElement, {}
            )
            di = 0

            def make_element(iv):
                i, v = iv
                elem = d_factory.new_descriptor('test', i)
                elem.set_vector(v)
                return elem

            # Constructing artificial descriptors
            x = numpy.random.rand(N, DIM)
            x_p1 = x[x[:, 1] <= 0.30]
            x_p2 = x[(x[:, 1] >= 0.36) & (x[:, 1] <= 0.63)]
            x_p3 = x[x[:, 1] >= 0.69]

            d_p1 = p.map(make_element, enumerate(x_p1, di))
            di += len(d_p1)
            d_p2 = p.map(make_element, enumerate(x_p2, di))
            di += len(d_p2)
            d_p3 = p.map(make_element, enumerate(x_p3, di))
            di += len(d_p3)

            # Create/Train test classifier
            classifier = LibSvmClassifier(
                train_params={
                    '-t': 0,  # linear kernel
                    '-b': 1,  # enable probability estimates
                    '-c': 2,  # SVM-C parameter C
                    '-q': ''  # quite mode
                },
                normalize=None,  # DO NOT normalize descriptors
            )
            classifier.train({P1_LABEL: d_p1, P2_LABEL: d_p2, P3_LABEL: d_p3})

            # Test classifier
            x = numpy.random.rand(N, DIM)
            x_p1 = x[x[:, 1] <= 0.30]
            x_p2 = x[(x[:, 1] >= 0.36) & (x[:, 1] <= 0.63)]
            x_p3 = x[x[:, 1] >= 0.69]

            d_p1 = p.map(make_element, enumerate(x_p1, di))
            di += len(d_p1)
            d_p2 = p.map(make_element, enumerate(x_p2, di))
            di += len(d_p2)
            d_p3 = p.map(make_element, enumerate(x_p3, di))
            di += len(d_p3)

            d_p1_sync = {}
            for d in d_p1:
                c = classifier.classify(d, c_factory)
                self.assertEqual(c.max_label(),
                                 P1_LABEL,
                                 "Incorrect %s label: %s :: %s" %
                                 (P1_LABEL, d.vector(), c.get_classification()))
                d_p1_sync[d] = c

            d_p2_sync = {}
            for d in d_p2:
                c = classifier.classify(d, c_factory)
                self.assertEqual(c.max_label(),
                                 P2_LABEL,
                                 "Incorrect %s label: %s :: %s" %
                                 (P2_LABEL, d.vector(), c.get_classification()))
                d_p2_sync[d] = c

            d_neg_sync = {}
            for d in d_p3:
                c = classifier.classify(d, c_factory)
                self.assertEqual(c.max_label(),
                                 P3_LABEL,
                                 "Incorrect %s label: %s :: %s" %
                                 (P3_LABEL, d.vector(), c.get_classification()))
                d_neg_sync[d] = c

            # test that async classify produces the same results
            # -- p1
            async_p1 = classifier.classify_async(d_p1, c_factory)
            self.assertEqual(async_p1, d_p1_sync,
                             "Async computation of p1 set did not yield "
                             "the same results as synchronous computation.")
            # -- p2
            async_p2 = classifier.classify_async(d_p2, c_factory)
            self.assertEqual(async_p2, d_p2_sync,
                             "Async computation of p2 set did not yield "
                             "the same results as synchronous computation.")
            # -- neg
            async_neg = classifier.classify_async(d_p3, c_factory)
            self.assertEqual(async_neg, d_neg_sync,
                             "Async computation of neg set did not yield "
                             "the same results as synchronous computation.")
            # -- combined -- threaded
            sync_combined = dict(six.iteritems(d_p1_sync))
            sync_combined.update(d_p2_sync)
            sync_combined.update(d_neg_sync)
            async_combined = classifier.classify_async(
                d_p1 + d_p2 + d_p3, c_factory,
                use_multiprocessing=False
            )
            self.assertEqual(async_combined, sync_combined,
                             "Async computation of all test descriptors "
                             "did not yield the same results as "
                             "synchronous classification.")
            # -- combined -- multiprocess
            async_combined = classifier.classify_async(
                d_p1 + d_p2 + d_p3, c_factory,
                use_multiprocessing=True
            )
            self.assertEqual(async_combined, sync_combined,
                             "Async computation of all test descriptors "
                             "(mixed order) did not yield the same results "
                             "as synchronous classification.")

            # Closing resources
            p.close()
            p.join()



from __future__ import division, print_function
import unittest

from smqtk.utils.url import (
    url_join
)


class TestUrlTools (unittest.TestCase):

    def test_url_join_simple(self):
        # One parameter
        self.assertEqual(url_join('foo'), 'foo')

        # multi-parameter, convert to str
        self.assertEqual(
            url_join('https://foo', 'bar', 1, 'six'),
            'https://foo/bar/1/six'
        )

    def test_url_join_empty_all(self):
        self.assertEqual(url_join(''), '')

        self.assertEqual(
            url_join('', ''),
            ''
        )

        self.assertEqual(
            url_join('', '', ''),
            ''
        )

    def test_url_join_empty_leading(self):
        self.assertEqual(
            url_join('', 'foo'),
            'foo'
        )

        self.assertEqual(
            url_join('', '', 'foo', 'bar'),
            'foo/bar'
        )

    def test_url_join_empty_middle(self):
        self.assertEqual(
            url_join('foo', '', 'bar'),
            'foo/bar'
        )

        self.assertEqual(
            url_join('foo', '', '', 'bar', '', 'baz'),
            'foo/bar/baz'
        )

    def test_url_join_empty_last(self):
        self.assertEqual(
            url_join('foo', ''),
            "foo/"
        )

    def test_url_join_empty_mixed(self):
        self.assertEqual(
            url_join('', '', 'b', '', 'c', '', '', 'a', '', ''),
            'b/c/a/'
        )

        self.assertEqual(
            url_join('', '', 'b', '', 'c', '', '', 'a'),
            'b/c/a'
        )

    def test_url_join_protocol_handling(self):
        self.assertEqual(
            url_join('http://'),
            'http://'
        )

        self.assertEqual(
            url_join('http://', 'https://'),
            'https://'
        )

        # not that this will probably ever be an intended result, this tests
        # documented logic.
        self.assertEqual(
            url_join('http://', 'https:/'),
            'http://https:'
        )

    def test_url_join_restart_protocol(self):
        # Test restarting url concat due to protocol header
        self.assertEqual(
            url_join('http://a.b.c', 'ftp://ba.c'),
            'ftp://ba.c'
        )

        self.assertEqual(
            url_join('', 'a', 'b', 'https://', 'bar', ''),
            'https://bar/'
        )

    def test_url_join_restart_slash(self):
        self.assertEqual(
            url_join("foo", '/bar', 'foo'),
            '/bar/foo'
        )

        self.assertEqual(
            url_join("foo", '/bar', '/foo'),
            '/foo'
        )

        self.assertEqual(
            url_join("foo", '/bar', '/'),
            '/'
        )

        self.assertEqual(
            url_join("foo", '/bar', '/', 'foo'),
            '/foo'
        )

    def test_url_join_restart_mixed(self):
        self.assertEqual(
            url_join("foo", '/bar', 'https://foo'),
            'https://foo'
        )

        self.assertEqual(
            url_join("foo", 'https://foo', '/bar'),
            '/bar'
        )



import csv
import errno
import os
import numpy
import re
import tempfile
import threading
import time

from smqtk.utils import SmqtkObject


def safe_create_dir(d):
    """
    Recursively create the given directory, ignoring the already-exists
    error if thrown.

    :param d: Directory filepath to create
    :type d: str

    :return: The directory that was created, i.e. the directory that was passed
        (in absolute form).
    :rtype: str

    """
    d = os.path.abspath(os.path.expanduser(d))
    try:
        os.makedirs(d)
    except OSError as ex:
        if ex.errno == errno.EEXIST and os.path.exists(d):
            pass
        else:
            raise
    return d


def safe_file_write(path, b, tmp_dir=None):
    """
    Safely write to a file in such a way that the target file is never
    incompletely written to due to error or multiple agents attempting to
    writes.

    We leverage that most OSs have an atomic move/rename operation by first
    writing bytes to a separate temporary file first, then renaming the
    temporary file to the final destination path when write is complete.
    Temporary files are written to the same directory as the target file unless
    otherwise specified.

    **NOTE:** *Windows does not have an atomic file rename and this function
    currently does not do anything special to ensure atomic rename on Windows.*

    :param path: Path to the file to write to.
    :type path: str

    :param b: Byte iterable to write to file.
    :type b: str | bytes

    :param tmp_dir: Optional custom directory to write the intermediate
        temporary file to. This directory must already exist.
    :type tmp_dir: None | str

    """
    file_dir = os.path.dirname(path)
    file_name = os.path.basename(path)
    file_base, file_ext = os.path.splitext(file_name)

    # Make sure containing directory exists
    safe_create_dir(file_dir)

    # Write to a temporary file first, then OS move the temp file to the final
    # destination. This is due to, on most OSs, a file rename/move being atomic.
    # TODO(paul.tunison): Do something else on windows since moves there are not
    #   guaranteed atomic.
    tmp_dir = file_dir if tmp_dir is None else tmp_dir
    f = tempfile.NamedTemporaryFile(suffix=file_ext, prefix=file_base + '.',
                                    dir=tmp_dir, delete=False)
    try:
        with f:
            f.write(b)
    except Exception:
        # Remove temporary file if anything bad happens.
        os.remove(f.name)
        raise
    os.rename(f.name, path)


def make_tempfile(suffix="", prefix="tmp", directory=None, text=False):
    """
    Wrapper for ``tempfile.mkstemp`` that closes/discards the file descriptor
    returned from the method. Arguments/keywords passed are the same as, and
    passed directly to ``tempfile.mkstemp``.

    :return: Path to a new user-owned temporary file.
    :rtype: str

    """
    fd, fp = tempfile.mkstemp(suffix, prefix, directory, text)
    os.close(fd)
    return fp


def iter_directory_files(d, recurse=True):
    """
    Iterates through files in the structure under the given directory.

    :param d: base directory path
    :type d: str

    :param recurse: If true, we recursively descend into all directories under
        the given directory. If false, we only return the files in the given
        directory and not the sub-directories in the given directory. If this is
        an integer (positive), on only recurse that many sub-directories.
    :type recurse: bool | int

    :return: Generator expression yielding absolute file paths under the given
        directory.
    :rtype: collections.Iterable[str]

    """
    d = os.path.abspath(d)
    for dirpath, dirnames, filenames in os.walk(d):
        for fname in filenames:
            yield os.path.join(dirpath, fname)
        if not recurse:
            break
        elif recurse is not True and dirpath != d:
            # must be an integer
            level = len(os.path.relpath(dirpath, d).split(os.sep))
            if level == recurse:
                # Empty directories descending
                del dirnames[:]
        # else recurse fully


def touch(fname):
    """
    Touch a file, creating it if it doesn't exist, setting its updated time to
    now.

    :param fname: File path to touch.
    :type fname: str

    """
    with open(fname, 'a'):
        os.utime(fname, None)


def exclusive_touch(file_path):
    """
    Attempt to touch a file. If that file already exists, we return False.
    If the file was touched and created, we return True. Other OSErrors
    thrown beside the expected "file already exists" error are passed
    upwards.

    :param file_path: Path to the file to touch.
    :type file_path: str

    :return: True if we touched/created the file, false if we couldn't
    :rtype: bool

    """
    try:
        fd = os.open(file_path, os.O_CREAT | os.O_EXCL)
        os.close(fd)
        return True
    except OSError as ex:
        if ex.errno == 17:  # File exists, could not touch.
            return False
        else:
            raise


def iter_svm_file(filepath, width):
    """
    Iterate parsed vectors in a parsed "*.svm" file that encodes a sparce
    matrix, where each line consists of multiple "index:value" pairs in index
    order. Multiple lines construct a matrix.

    :param filepath: Path to the SVM file encoding an array per line
    :type filepath: str
    :param width: The known number of columns in the sparse vectors.
    :param width: int

    :return: Generator yielding ndarray vectors
    :rtype: collections.Iterable[numpy.core.multiarray.ndarray]

    """
    idx_val_re = re.compile("([0-9]+):([-+]?[0-9]*\.?[0-9]*)")
    with open(filepath, 'r') as infile:
        for line in infile:
            v = numpy.zeros(width, dtype=float)
            for seg in line.strip().split(' '):
                m = idx_val_re.match(seg)
                assert m is not None, \
                    "Invalid index:value match for segment '%s'" % seg
                idx, val = int(m.group(1)), float(m.group(2))
                v[idx] = val
            yield v


def iter_csv_file(filepath):
    """
    Iterate parsed vectors in a "*.csv" file that encodes descriptor output
    where each line is a descriptor vector. Multiple lines construct a matrix.

    :param filepath: Path to the CSV file encoding an array per line.
    :type filepath: str

    :return: Generator yielding ndarray vectors
    :rtype: collections.Iterable[numpy.core.multiarray.ndarray]

    """
    with open(filepath) as f:
        r = csv.reader(f)
        for l in r:
            yield numpy.array(l, dtype=float)


def file_mimetype_filemagic(filepath):
    """
    Determine file mimetype using the file-magic module.

    The file the given path refers to must exist.

    :raises ImportError: ``magic`` python module not available.
    :raises IOError: ``filepath`` did not refer to an existing file.

    :param filepath: Path to the (existing) file to determine the mimetype of.
    :type filepath: str

    :return: MIMETYPE string identifier.
    :rtype: str

    """
    # noinspection PyUnresolvedReferences
    import magic
    if os.path.isfile(filepath):
        d = magic.detect_from_filename(filepath)
        return d.mime_type
    elif os.path.isdir(filepath):
        raise IOError(21, "Is a directory: '%s'" % filepath)
    else:
        raise IOError(2, "No such file or directory: '%s'" % filepath)


def file_mimetype_tika(filepath):
    """
    Determine file mimetype using ``tika`` module.

    The file the given path refers to must exist. This function may fail under
    multiprocessing situations.

    :raises ImportError: ``tika`` python module not available.
    :raises IOError: ``filepath`` did not refer to an existing file.

    :param filepath: Path to the (existing) file to determine the mimetype of.
    :type filepath: str

    :return: MIMETYPE string identifier.
    :rtype: str

    """
    # noinspection PyUnresolvedReferences
    import tika.detector
    return tika.detector.from_file(filepath)


class FileModificationMonitor (SmqtkObject, threading.Thread):
    """
    Utility object for triggering a callback function when an observed file
    changes based on file modification times observed
    """

    STATE_WAITING = 0   # Waiting for file to be modified
    STATE_WATCHING = 1  # Waiting for file to settle
    STATE_SETTLED = 2   # File has stopped being modified for settle period

    def __init__(self, filepath, monitor_interval, settle_window, callback):
        """
        On a separate thread, monitor the modification time of the file at the
        given file path. When the file is updated (after the file has stopped
        changing), trigger the provided callback function, given the monitored
        file path and the file stat event.

        We initially set ourselves as a daemon as that is the most probable
        usage of our functionality.

        :param filepath: Path to the file to monitor
        :type filepath: str

        :param monitor_interval: Frequency in seconds at which we check file
            modification times. This must be >= 0.
        :type monitor_interval: float

        :param settle_window: If a recently modified file is not modified again
            for this many seconds in a row, we consider the file done being
            modified and trigger the triggers ``callback``. This must be >= 0
            and should be >= the ``monitor_interval``.
        :type settle_window: float

        :param callback: Callback function that will be triggered every time
            the provided file has been updated and the settle time has safely
            expired.
        :type callback: (str) -> None

        :raises ValueError: The given filepath did not point to an existing,
            valid file.

        """
        SmqtkObject.__init__(self)
        threading.Thread.__init__(self, name=self.__class__.__name__)

        self.daemon = True

        self.filepath = filepath
        self.monitor_interval = monitor_interval
        self.settle_window = settle_window
        self.callback = callback

        self.event_stop = threading.Event()
        self.event_stop.set()  # make sure false

        self.state = self.STATE_WAITING

        if not os.path.isfile(self.filepath):
            raise ValueError("Provided filepath did not point to an existing, "
                             "valid file.")

        if monitor_interval < 0 or settle_window < 0:
            raise ValueError("Monitor and settle times must be >= 0")

    def stop(self):
        self._log.debug("stopped externally")
        self.event_stop.set()

    def stopped(self):
        return self.event_stop.is_set()

    def start(self):
        # Clear stop flag
        self.event_stop.clear()
        super(FileModificationMonitor, self).start()

    def run(self):
        # self._log.debug("starting run method")
        # mtime baseline
        last_mtime = os.path.getmtime(self.filepath)

        try:
            while not self.stopped():
                mtime = os.path.getmtime(self.filepath)

                # file has been updated
                if self.state == self.STATE_WAITING and last_mtime != mtime:
                    self.state = self.STATE_WATCHING
                    self._log.debug('change detected '
                                    '(mtime: %f -> %f, diff=%f) '
                                    ':: state(WAITING -> WATCHING)',
                                    last_mtime, mtime, mtime - last_mtime)

                # Wait until file is not being modified any more
                elif self.state == self.STATE_WATCHING:
                    t = time.time()
                    if t - mtime >= self.settle_window:
                        self.state = self.STATE_SETTLED
                        self._log.debug('file settled '
                                        '(mtime=%f, t=%f, diff=%f) '
                                        ':: state(WATCHING -> SETTLED)',
                                        mtime, t, t - mtime)
                    else:
                        self._log.debug('waiting for settle '
                                        '(mtime=%f, t=%f, diff=%f)...',
                                        mtime, t, t - mtime)
                        time.sleep(self.monitor_interval)

                elif self.state == self.STATE_SETTLED:
                    self.callback(self.filepath)
                    self.state = self.STATE_WAITING
                    self._log.debug('calling callback '
                                    ':: state(SETTLED -> WAITING)')

                # waiting for modification
                else:
                    # self._log.debug("waiting...")
                    time.sleep(self.monitor_interval)

                last_mtime = mtime
        finally:
            self.event_stop.set()

        self._log.debug('exiting')



"""
Retrieve the nearest neighbors of a set of UUIDs.

Input UIDs, specified in files input to the -u/--uuid-list option, are located
within the configured "descriptor_set" and then queried against the configured
"nn_index".  Thus, all UIDs provided in the input file should exist in the
configured "descriptor_set" otherwise a KeyError is raised.  Input descriptors
do not have to be necessarily indexed in the configured "nn_index".
"""
import logging
import os
import sys
from six.moves import zip

from smqtk.utils.bin_utils import (
    basic_cli_parser,
    utility_main_helper,
)

from smqtk.algorithms.nn_index import get_nn_index_impls
from smqtk.representation import get_descriptor_index_impls
from smqtk.utils import plugin


__author__ = 'dan.lamanna@kitware.com'


def get_cli_parser():
    parser = basic_cli_parser(__doc__)
    parser.add_argument('-u', '--uuid-list',
                        default=None, metavar='PATH',
                        help='Path to list of UUIDs to calculate nearest '
                             'neighbors for. If empty, all UUIDs present '
                             'in the descriptor index will be used.')

    parser.add_argument('-n', '--num',
                        default=10, metavar='INT', type=int,
                        help='Number of maximum nearest neighbors to return '
                             'for each UUID, defaults to retrieving 10 nearest '
                             'neighbors. Set to 0 to retrieve all nearest '
                             'neighbors.')
    return parser


def get_default_config():
    return {
        'plugins': {
            'descriptor_set': plugin.make_config(get_descriptor_index_impls()),
            'nn_index': plugin.make_config(get_nn_index_impls())
        }
    }


def main():
    # Print help and exit if no arguments were passed
    if len(sys.argv) == 1:
        get_cli_parser().print_help()
        sys.exit(1)

    args = get_cli_parser().parse_args()
    config = utility_main_helper(get_default_config, args)

    log = logging.getLogger(__name__)
    log.debug('Showing debug messages.')

    #: :type: smqtk.representation.DescriptorIndex
    descriptor_set = plugin.from_plugin_config(
        config['plugins']['descriptor_set'], get_descriptor_index_impls()
    )
    #: :type: smqtk.algorithms.NearestNeighborsIndex
    nearest_neighbor_index = plugin.from_plugin_config(
        config['plugins']['nn_index'], get_nn_index_impls()
    )

    # noinspection PyShadowingNames
    def nearest_neighbors(descriptor, n):
        if n == 0:
            n = len(nearest_neighbor_index)

        uuids, descriptors = nearest_neighbor_index.nn(descriptor, n)
        # Strip first result (itself) and create list of (uuid, distance)
        return list(zip([x.uuid() for x in uuids[1:]], descriptors[1:]))

    if args.uuid_list is not None and not os.path.exists(args.uuid_list):
        log.error('Invalid file list path: %s', args.uuid_list)
        exit(103)
    elif args.num < 0:
        log.error('Number of nearest neighbors must be >= 0')
        exit(105)

    if args.uuid_list is not None:
        with open(args.uuid_list, 'r') as infile:
            for line in infile:
                descriptor = descriptor_set.get_descriptor(line.strip())
                print(descriptor.uuid())
                for neighbor in nearest_neighbors(descriptor, args.num):
                    print('%s,%f' % neighbor)
    else:
        for (uuid, descriptor) in descriptor_set.iteritems():
            print(uuid)
            for neighbor in nearest_neighbors(descriptor, args.num):
                print('%s,%f' % neighbor)


if __name__ == '__main__':
    main()



import six
from six.moves import cPickle as pickle

from smqtk.representation import DescriptorIndex, get_data_element_impls
from smqtk.utils import merge_dict, plugin, SimpleTimer


class MemoryDescriptorIndex (DescriptorIndex):
    """
    In-memory descriptor index with file caching.

    Stored descriptor elements are all held in memory in a uuid-to-element
    dictionary (hash table).

    If the path to a file cache is provided, it is loaded at construction if it
    exists. When elements are added to the index, the in-memory table is dumped
    to the cache.
    """

    @classmethod
    def is_usable(cls):
        """
        Check whether this class is available for use.

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        # no dependencies
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        turning those argument names into configuration dictionary keys. If any
        of those arguments have defaults, we will add those values into the
        configuration dictionary appropriately. The dictionary returned should
        only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(MemoryDescriptorIndex, cls).get_default_config()
        c['cache_element'] = plugin.make_config(get_data_element_impls())
        return c

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: MemoryDescriptorIndex

        """
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        # Optionally construct cache element from sub-config.
        if config_dict['cache_element'] \
                and config_dict['cache_element']['type']:
            e = plugin.from_plugin_config(config_dict['cache_element'],
                                          get_data_element_impls())
            config_dict['cache_element'] = e
        else:
            config_dict['cache_element'] = None

        return super(MemoryDescriptorIndex, cls).from_config(config_dict, False)

    def __init__(self, cache_element=None, pickle_protocol=-1):
        """
        Initialize a new in-memory descriptor index, or reload one from a
        cache.

        :param cache_element: Optional data element cache, loading an existing
            index if the element has bytes. If the given element is writable,
             new descriptors added to this index are cached to the element.
        :type cache_element: None | smqtk.representation.DataElement

        :param pickle_protocol: Pickling protocol to use when serializing index
            table to the optionally provided, writable cache element. We will
            use -1 by default (latest version, probably a binary form).
        :type pickle_protocol: int

        """
        super(MemoryDescriptorIndex, self).__init__()

        # Mapping of descriptor UUID to the DescriptorElement instance.
        #: :type: dict[collections.Hashable, smqtk.representation.DescriptorElement]
        self._table = {}
        # Record of optional file cache we're using
        self.cache_element = cache_element
        self.pickle_protocol = pickle_protocol

        if cache_element and not cache_element.is_empty():
            self._log.debug("Loading cached descriptor index table from %s "
                            "element.", cache_element.__class__.__name__)
            self._table = pickle.loads(cache_element.get_bytes())

    def get_config(self):
        c = merge_dict(self.get_default_config(), {
            "pickle_protocol": self.pickle_protocol,
        })
        if self.cache_element:
            merge_dict(c['cache_element'],
                       plugin.to_plugin_config(self.cache_element))
        return c

    def cache_table(self):
        if self.cache_element and self.cache_element.writable():
            with SimpleTimer("Caching descriptor table", self._log.debug):
                self.cache_element.set_bytes(pickle.dumps(self._table,
                                                          self.pickle_protocol))

    def count(self):
        return len(self._table)

    def clear(self):
        """
        Clear this descriptor index's entries.
        """
        self._table = {}
        self.cache_table()

    def has_descriptor(self, uuid):
        """
        Check if a DescriptorElement with the given UUID exists in this index.

        :param uuid: UUID to query for
        :type uuid: collections.Hashable

        :return: True if a DescriptorElement with the given UUID exists in this
            index, or False if not.
        :rtype: bool

        """
        return uuid in self._table

    def add_descriptor(self, descriptor, no_cache=False):
        """
        Add a descriptor to this index.

        Adding the same descriptor multiple times should not add multiple
        copies of the descriptor in the index.

        :param descriptor: Descriptor to index.
        :type descriptor: smqtk.representation.DescriptorElement

        :param no_cache: Do not cache the internal table if a file cache was
            provided. This would be used if adding many descriptors at a time,
            preventing a file write for every individual descriptor added.
        :type no_cache: bool

        """
        self._table[descriptor.uuid()] = descriptor
        if not no_cache:
            self.cache_table()

    def add_many_descriptors(self, descriptors):
        """
        Add multiple descriptors at one time.

        :param descriptors: Iterable of descriptor instances to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        added_something = False
        for d in descriptors:
            # using no-cache so we don't trigger multiple file writes
            self.add_descriptor(d, no_cache=True)
            added_something = True
        if added_something:
            self.cache_table()

    def get_descriptor(self, uuid):
        """
        Get the descriptor in this index that is associated with the given UUID.

        :param uuid: UUID of the DescriptorElement to get.
        :type uuid: collections.Hashable

        :raises KeyError: The given UUID doesn't associate to a
            DescriptorElement in this index.

        :return: DescriptorElement associated with the queried UUID.
        :rtype: smqtk.representation.DescriptorElement

        """
        return self._table[uuid]

    def get_many_descriptors(self, uuids):
        """
        Get an iterator over descriptors associated to given descriptor UUIDs.

        :param uuids: Iterable of descriptor UUIDs to query for.
        :type uuids: collections.Iterable[collections.Hashable]

        :raises KeyError: A given UUID doesn't associate with a
            DescriptorElement in this index.

        :return: Iterator of descriptors associated to given uuid values.
        :rtype: __generator[smqtk.representation.DescriptorElement]

        """
        for uid in uuids:
            yield self._table[uid]

    def remove_descriptor(self, uuid, no_cache=False):
        """
        Remove a descriptor from this index by the given UUID.

        :param uuid: UUID of the DescriptorElement to remove.
        :type uuid: collections.Hashable

        :raises KeyError: The given UUID doesn't associate to a
            DescriptorElement in this index.

        :param no_cache: Do not cache the internal table if a file cache was
            provided. This would be used if adding many descriptors at a time,
            preventing a file write for every individual descriptor added.
        :type no_cache: bool

        """
        del self._table[uuid]
        if not no_cache:
            self.cache_table()

    def remove_many_descriptors(self, uuids):
        """
        Remove descriptors associated to given descriptor UUIDs from this
        index.

        :param uuids: Iterable of descriptor UUIDs to remove.
        :type uuids: collections.Iterable[collections.Hashable]

        :raises KeyError: A given UUID doesn't associate with a
            DescriptorElement in this index.

        """
        for uid in uuids:
            # using no-cache so we don't trigger multiple file writes
            self.remove_descriptor(uid, no_cache=True)
        self.cache_table()

    def iterkeys(self):
        return six.iterkeys(self._table)

    def iterdescriptors(self):
        return six.itervalues(self._table)

    def iteritems(self):
        return six.iteritems(self._table)


DESCRIPTOR_INDEX_CLASS = MemoryDescriptorIndex



"""
Runs conforming SMQTK Web Applications.
"""

import logging

from flask_basicauth import BasicAuth
import six

from smqtk.utils import bin_utils
import smqtk.web


def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)

    # Application options
    group_application = parser.add_argument_group("Application Selection")
    group_application.add_argument('-l', '--list',
                                   default=False, action="store_true",
                                   help="List currently available applications "
                                        "for running. More description is "
                                        "included if SMQTK verbosity is "
                                        "increased (-v | --debug-smqtk)")
    group_application.add_argument('-a', '--application', default=None,
                                   help="Label of the web application to run.")

    # Server options
    group_server = parser.add_argument_group("Server options")
    group_server.add_argument('-r', '--reload',
                              action='store_true', default=False,
                              help='Turn on server reloading.')
    group_server.add_argument('-t', '--threaded',
                              action='store_true', default=False,
                              help="Turn on server multi-threading.")
    group_server.add_argument('--host',
                              default=None,
                              help="Run host address specification override. "
                                   "This will override all other configuration "
                                   "method specifications.")
    group_server.add_argument('--port',
                              default=None,
                              help="Run port specification override. This will "
                                   "override all other configuration method "
                                   "specifications.")
    group_server.add_argument("--use-basic-auth",
                              action="store_true", default=False,
                              help="Use global basic authentication as "
                                   "configured.")

    # Other options
    group_other = parser.add_argument_group("Other options")
    group_other.add_argument('--debug-server',
                             action='store_true', default=False,
                             help='Turn on server debugging messages ONLY')
    group_other.add_argument('--debug-smqtk',
                             action='store_true', default=False,
                             help='Turn on SMQTK debugging messages ONLY')

    return parser


def main():
    parser = cli_parser()
    args = parser.parse_args()

    debug_smqtk = args.debug_smqtk or args.verbose
    debug_server = args.debug_server or args.verbose

    bin_utils.initialize_logging(logging.getLogger("__main__"),
                                 logging.INFO - (10 * debug_smqtk))
    bin_utils.initialize_logging(logging.getLogger("smqtk"),
                                 logging.INFO - (10*debug_smqtk))
    bin_utils.initialize_logging(logging.getLogger("werkzeug"),
                                 logging.WARN - (20*debug_server))
    log = logging.getLogger(__name__)

    web_applications = smqtk.web.get_web_applications()

    if args.list:
        log.info("")
        log.info("Available applications:")
        log.info("")
        for l, cls in six.iteritems(web_applications):
            log.info("\t" + l)
            if debug_smqtk:
                log.info('\t' + ('^'*len(l)) + '\n' +

                         cls.__doc__ + '\n' +
                         ('*' * 80) + '\n')
        log.info("")
        exit(0)

    application_name = args.application

    if application_name is None:
        log.error("No application name given!")
        exit(1)
    elif application_name not in web_applications:
        log.error("Invalid application label '%s'", application_name)
        exit(1)

    #: :type: smqtk.web.SmqtkWebApp
    app_class = web_applications[application_name]

    config = bin_utils.utility_main_helper(app_class.get_default_config, args,
                                           skip_logging_init=True)

    host = args.host
    port = args.port and int(args.port)
    use_reloader = args.reload
    use_threading = args.threaded
    use_basic_auth = args.use_basic_auth

    # noinspection PyUnresolvedReferences
    #: :type: smqtk.web.SmqtkWebApp
    app = app_class.from_config(config)
    if use_basic_auth:
        app.config["BASIC_AUTH_FORCE"] = True
        BasicAuth(app)
    app.config['DEBUG'] = debug_server

    log.info("Starting application")
    app.run(host=host, port=port, debug=debug_server, use_reloader=use_reloader,
            threaded=use_threading)


if __name__ == "__main__":
    main()



import logging
import multiprocessing
import multiprocessing.queues
import sys
import threading
import time

import numpy
from six import next
from six.moves import queue

from smqtk.utils import SmqtkObject


__author__ = 'paul.tunison@kitware.com'


__all__ = [
    'elements_to_matrix',
]


def elements_to_matrix(descr_elements, mat=None, procs=None, buffer_factor=2,
                       report_interval=None, use_multiprocessing=False,
                       thread_q_put_interval=0.001):
    """
    Add to or create a numpy matrix, adding to it the vector data contained in
    a sequence of DescriptorElement instances using asynchronous processing.

    If ``mat`` is provided, its shape must equal:
        ( len(descr_elements) , descr_elements[0].size )

    :param descr_elements: Sequence of DescriptorElement objects to transform
        into a matrix. Each element should contain descriptor vectors of the
        same size.
    :type descr_elements:
        collections.Sequence[smqtk.representation.DescriptorElement] |
        collections.Iterable[smqtk.representation.DescriptorElement]

    :param mat: Optionally a pre-constructed numpy matrix of the shape
        ``(nDescriptors, nFeatures)`` to load descriptor vectors into. We will
        only iterate ``nDescriptors`` into the given ``descr_elements``
        iterable. If there are more rows in the given matrix than there are
        DescriptorElements in ``descr_elements``, then not all rows in the
        given matrix will be set. Elements yielded by ``descr_elements`` must
        be of the same dimensionality as this given matrix (``nFeatures``)
        otherwise an exception will be raised (``ValueError``, by numpy).

        If this is not supplied, we create a new matrix to insert vectors into
        based on the number of input descriptor elements. This mode required
        that the input elements are in a container that defines __len__
    :type mat: None | numpy.ndarray

    :param procs: Optional specification of the number of threads/cores to use.
        If None, we will attempt to use all available threads/cores.
    :type procs: None | int | long

    :param buffer_factor: Multiplier against the number of processes used to
        limit the growth size of the result queue coming from worker processes.
    :type buffer_factor: float

    :param report_interval: Optional interval in seconds for debug logging to
        occur reporting about conversion speed. This should be greater than 0
        if this debug logging is desired.
    :type report_interval: None | float

    :param use_multiprocessing: Whether or not to use discrete processes as the
        parallelization agent vs python threads.
    :type use_multiprocessing: bool

    :param thread_q_put_interval: Interval at worker threads attempt to insert
        values into the output queue after fetching vector from a
        DescriptorElement. This is for dead-lock protection due to size-limited
        output queue. This is only used if ``use_multiprocessing`` is ``False``
        and this must be >0.
    :type thread_q_put_interval: float

    :return: Created or input matrix.
    :rtype: numpy.ndarray

    """
    log = logging.getLogger(__name__)

    # Create/check matrix
    if mat is None:
        sample = next(iter(descr_elements))
        sample_v = sample.vector()
        shp = (len(descr_elements),
               sample_v.size)
        log.debug("Creating new matrix with shape: %s", shp)
        mat = numpy.ndarray(shp, sample_v.dtype)

    if procs is None:
        procs = multiprocessing.cpu_count()

    # Choose parallel types
    worker_kwds = {}
    if use_multiprocessing:
        queue_t = multiprocessing.Queue
        worker_t = _ElemVectorExtractorProcess
    else:
        queue_t = queue.Queue
        worker_t = _ElemVectorExtractorThread

        assert thread_q_put_interval >= 0, \
            "Thread queue.put interval must be >= 0. (given: %f)" \
            % thread_q_put_interval
        worker_kwds['q_put_interval'] = thread_q_put_interval

    in_q = queue_t()
    out_q = queue_t(int(procs * buffer_factor))
    # Workers for async extraction
    log.debug("constructing worker processes")
    workers = [worker_t(i, in_q, out_q, **worker_kwds) for i in range(procs)]

    in_queue_t = _FeedQueueThread(descr_elements, in_q, mat, len(workers))

    try:
        # Start worker processes
        log.debug("starting worker processes")
        for w in workers:
            w.daemon = True
            w.start()

        log.debug("Sending work packets")
        in_queue_t.daemon = True
        in_queue_t.start()

        # Collect work from async
        log.debug("Aggregating async results")
        terminals_collected = 0
        f = 0
        lt = t = time.time()
        while terminals_collected < len(workers):
            packet = out_q.get()
            if packet is None:
                terminals_collected += 1
            elif isinstance(packet, Exception):
                raise packet
            else:
                r, v = packet
                mat[r] = v

                f += 1
                if report_interval and time.time() - lt >= report_interval:
                    log.debug("Rows per second: %f, Total: %d",
                              f / (time.time() - t), f)
                    lt = time.time()

        # All work should be exhausted at this point
        if use_multiprocessing and sys.platform == 'darwin':
            # multiprocessing.Queue.qsize doesn't work on OSX
            # Try to get something from each queue, expecting an empty exception
            try:
                in_q.get(block=False)
            except multiprocessing.queues.Empty:
                pass
            else:
                raise AssertionError("In queue not empty")
            try:
                out_q.get(block=False)
            except multiprocessing.queues.Empty:
                pass
            else:
                raise AssertionError("Out queue not empty")
        else:
            assert in_q.qsize() == 0, "In queue not empty"
            assert out_q.qsize() == 0, "Out queue not empty"

        return mat
    finally:
        log.debug("Stopping/Joining queue feeder thread")
        in_queue_t.stop()
        in_queue_t.join()

        if use_multiprocessing:
            # Forcibly terminate worker processes if still alive
            log.debug("Joining/Terminating process workers")
            for w in workers:
                if w.is_alive():
                    w.terminate()
                w.join()

            log.debug("Cleaning multiprocess queues")
            for q in (in_q, out_q):
                q.close()
                q.join_thread()
        else:
            log.debug("Stopping/Joining threaded workers")
            for w in workers:
                w.stop()
                # w.join()
                # Threads should exit fine from here

        log.debug("Done")


class _FeedQueueThread (SmqtkObject, threading.Thread):

    def __init__(self, descr_elements, q, out_mat, num_terminal_packets):
        super(_FeedQueueThread, self).__init__()

        self.num_terminal_packets = num_terminal_packets
        self.out_mat = out_mat
        self.q = q
        self.descr_elements = descr_elements

        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()

    def run(self):
        try:
            # Special case for in-memory storage of descriptors
            from smqtk.representation.descriptor_element.local_elements \
                import DescriptorMemoryElement

            for r, d in enumerate(self.descr_elements):
                # If we've run out of matrix to fill,
                if r >= self.out_mat.shape[0]:
                    break

                if isinstance(d, DescriptorMemoryElement):
                    self.out_mat[r] = d.vector()
                else:
                    self.q.put((r, d))

                # If we're told to stop, immediately quit out of processing
                if self.stopped():
                    break
        except KeyboardInterrupt:
            self._log.debug("FeedQueueThread received KeyboardInterrupt")
        except Exception as ex:
            self._log.error("Feeder thread encountered an exception: %s",
                            str(ex))
            self.q.put(ex)
        finally:
            self._log.debug("Sending in-queue terminal packets")
            for _ in range(self.num_terminal_packets):
                self.q.put(None)
            self._log.debug("Closing in-queue")


class _ElemVectorExtractorProcess (SmqtkObject, multiprocessing.Process):
    """
    Helper process for extracting DescriptorElement vectors on a separate
    process. This terminates with a None packet fed to in_q. Otherwise, in_q
    values are expected to be (row, element) pairs. Tuples of the form
    (row, vector) are published to the out_q.

    Terminal value: None

    """

    def __init__(self, i, in_q, out_q):
        super(_ElemVectorExtractorProcess, self)\
            .__init__(name='[w%d]' % i)
        self._log.debug("Making process worker (%d, %s, %s)", i, in_q, out_q)
        self.i = i
        self.in_q = in_q
        self.out_q = out_q

    def run(self):
        try:
            packet = self.in_q.get()
            while packet is not None:
                if isinstance(packet, Exception):
                    self.out_q.put(packet)
                else:
                    row, elem = packet
                    v = elem.vector()
                    self.out_q.put((row, v))
                packet = self.in_q.get()
            self.out_q.put(None)
        except KeyboardInterrupt:
            self._log.debug("ElemVectorExtractorProcess received "
                            "KeyboardInterrupt")
        except Exception as ex:
            self._log.error("%s%s encountered an exception: %s",
                            self.__class__.__name__, self.name,
                            str(ex))
            self.out_q.put(ex)


class _ElemVectorExtractorThread (SmqtkObject, threading.Thread):
    """
    Helper process for extracting DescriptorElement vectors on a separate
    process. This terminates with a None packet fed to in_q. Otherwise, in_q
    values are expected to be (row, element) pairs. Tuples of the form
    (row, vector) are published to the out_q.

    Terminal value: None

    """

    def __init__(self, i, in_q, out_q, q_put_interval=0.001):
        SmqtkObject.__init__(self)
        threading.Thread.__init__(self, name='[w%d]' % i)

        self._log.debug("Making thread worker (%d, %s, %s)", i, in_q, out_q)
        self.i = i
        self.in_q = in_q
        self.out_q = out_q
        self.q_put_interval = q_put_interval

        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()

    def run(self):
        try:
            packet = self.in_q.get()
            while packet is not None and not self.stopped():
                if isinstance(packet, Exception):
                    self.out_q.put(packet)
                else:
                    row, elem = packet
                    v = elem.vector()
                    self.q_put((row, v))
                packet = self.in_q.get()
            self.q_put(None)
        except KeyboardInterrupt:
            self._log.debug("ElemVectorExtractorThread received "
                            "KeyboardInterrupt")
        except Exception as ex:
            self._log.error("%s%s encountered an exception: %s",
                            self.__class__.__name__, self.name,
                            str(ex))
            self.out_q.put(ex)

    def q_put(self, val):
        """
        Try to put the given value into the output queue until it is inserted
        (if it was previously full), or the stop signal was given.
        """
        put = False
        while not put and not self.stopped():
            try:
                self.out_q.put(val, timeout=self.q_put_interval)
                put = True
            except queue.Full:
                # self._log.debug("Skipping q.put Full error")
                pass



"""
Utility functions for colorDescriptor operations
"""

import logging
import numpy
import os
import os.path as osp
# noinspection PyPackageRequirements
import PIL.Image
import subprocess
import sys
import tempfile

# Try to import DescriptorIO module provided by ColorDescriptor package
try:
    import DescriptorIO
except ImportError:
    try:
        from colorDescriptor import DescriptorIO
    except ImportError:
        DescriptorIO = None


def has_colordescriptor_module():
    """
    :return: Boolean describing whether the required colorDescriptor module was
        found or not. If not found, the base descriptor functionality does not
        exist.
    :rtype: bool
    """
    return DescriptorIO is not None


def generate_descriptors(cd_exe, img_filepath, descriptor_type,
                         info_matrix_path, descr_matrix_path,
                         limit_descriptors=None,
                         recompute=False):
    """
    Execute the given colorDescriptor executable, saving the generated info
    and descriptor matrices for the provided image to the provided file paths.
    Descriptor matrix is normalized into histograms of relative frequencies
    instead of histograms of raw bin counts.

    This does NOT return matrices directly due to memory concerns, especially
    in regards to multiprocessing as multiple copies of the matrix exist in the
    system, leading to excessive memory clogging.

    Matrices are saved in numpy binary format (.npy). ``numpy.load`` function
    should be used to load matrices back in.

    :raises ImportError: The required python module for colorDescriptor IO is
        not available.
    :raises RuntimeError: Failed to generate output files or matrices for the '
        given input.

    :param cd_exe: ColorDescriptor executable to use
    :type cd_exe: str

    :param img_filepath: Path to the image file to process
    :type img_filepath: str

    :param descriptor_type: String type of descriptor to use from
        colorDescriptor.
    :type descriptor_type: str

    :param info_matrix_path: Path to where the computed information matrix
        for the given file should be saved. This will be saved as a numpy
        binary file (.npy).
    :type info_matrix_path: str

    :param descr_matrix_path: Path to where the computed descriptor matrix
        for the given file should be saved. This will be saved as a numpy
        binary file (.npy).
    :type descr_matrix_path: str

    :param limit_descriptors: Limit the number of descriptors generated if we
        were to produce more than the limit. If we exceed the limit, we randomly
        subsample down to the limit.
    :type limit_descriptors: int

    :param recompute: Force re-computation of descriptors for the given image
        file. This causes possible existing output files to be overwritten.
    :type recompute: bool

    :return: Shape information for info and descriptor matrices
    :rtype: ((int, int), (int, int))

    """
    if not has_colordescriptor_module():
        raise ImportError("Cannot find the DescriptorIO module provided by "
                          "ColorDescriptor. Read the README for dependencies!")

    log = logging.getLogger("ColorDescriptor::generate_descriptors{%s,%s}"
                            % (descriptor_type, osp.basename(img_filepath)))

    if not recompute \
            and osp.isfile(info_matrix_path) \
            and osp.isfile(descr_matrix_path):
        # log.debug("Found existing matrix files, loading shapes.")
        return (numpy.load(info_matrix_path).shape,
                numpy.load(descr_matrix_path).shape)

    # Determine the spacing between sample points in the image. We want have at
    # least 50 sample points along the shortest side with a minimum of 6 pixels
    # distance between sample points.
    try:
        w, h = PIL.Image.open(img_filepath).size
    except IOError as ex:
        raise RuntimeError("Could not open image at filepath '%s': %s"
                           % (img_filepath, str(ex)))
    ds_spacing = max(int(min(w, h) / 50.0), 6)
    log.debug("dense-sample spacing: %d", ds_spacing)

    tmp_fd, tmp_path = tempfile.mkstemp(prefix='colorDescriptor.')
    os.close(tmp_fd)

    log.debug("launching executable subprocess")
    # TODO: Perform harrislaplace detection method, if yields 0 descriptors, run
    #       densesample method. When harrislaplace fails, this generally means
    #       that there are no edge features in the image, e.g. a solid color
    #       image.
    cmd = [cd_exe, img_filepath, '--output', tmp_path,
           # "harrislaplace" is another option here, but this can result in 0
           # descriptors for an image, and is also slower.
           '--detector', 'densesampling',
           # '--detector', 'harrislaplace',
           '--ds_spacing', str(ds_spacing),
           '--descriptor', descriptor_type]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.communicate()

    # Info matrix consists of [x, y, scale, orientation, corner-ness]
    # - See colorDescriptor documentation for more information
    try:
        log.debug("Reading descriptors output")
        info, descriptors = DescriptorIO.readDescriptors(tmp_path)
    except IOError as ex:
        raise RuntimeError("ColorDescriptor failed to generate proper output "
                           "file. See error log for details. (error: %s"
                           % str(ex))
    finally:
        os.remove(tmp_path)

    # Also error if the descriptor is empty
    if not descriptors.shape[1]:
        raise RuntimeError("Produced empty descriptor.")

    # Divides each row in the descriptors matrix with the row-wise sum.
    # - This results in histograms for relative frequencies instead of direct
    #   bin counts.
    # - Adding float_info.min to row sums to prevent div-by-zero exception while
    #   introducing minimal numerical error.
    # noinspection PyUnresolvedReferences
    log.debug("normalizing histogram into relative frequency")
    # noinspection PyUnresolvedReferences
    descriptors = descriptors / (numpy.matrix(descriptors).sum(axis=1) +
                                 sys.float_info.min).A

    # Randomly sample rows down to this count if what was generated exceeded the
    # limit.
    if limit_descriptors and info.shape[0] > limit_descriptors:
        idxs = numpy.random.permutation(numpy.arange(info.shape[0]))
        idxs = sorted(idxs[:limit_descriptors])
        info = info[idxs, :]
        descriptors = descriptors[idxs, :]

    numpy.save(info_matrix_path, info)
    numpy.save(descr_matrix_path, descriptors)
    return info.shape, descriptors.shape



from __future__ import print_function

import atexit
import os
import tempfile
import time
import threading
import unittest

import smqtk.utils.file_utils


class TestFileModificationMonitor (unittest.TestCase):

    @staticmethod
    def _mk_test_fp():
        fd, fp = tempfile.mkstemp()
        os.close(fd)
        atexit.register(lambda: os.remove(fp))
        return fp

    def test_monitor_stop(self):
        # test that monitor stops when its told to
        fp = self._mk_test_fp()

        has_triggered = [False]

        def cb(filepath):
            has_triggered[0] = True
            self.assertEqual(filepath, fp)

        interval = 0.01
        monitor = smqtk.utils.file_utils.FileModificationMonitor(fp, interval,
                                                                 0.5, cb)
        self.assertTrue(monitor.stopped())

        monitor.start()

        try:
            self.assertFalse(has_triggered[0])
            self.assertTrue(monitor.is_alive())
            self.assertFalse(monitor.stopped())

            monitor.stop()
            # If thread hasn't entered while loop yet, it will immediately kick
            # out, otherwise its sleeping for the given interval.
            monitor.join(interval*2)

            self.assertFalse(has_triggered[0])
            self.assertFalse(monitor.is_alive())
        finally:
            if monitor.is_alive():
                print("WARNING :: Forcing thread stop by removing filepath var")
                monitor.filepath = None

    def test_short_file_copy(self):
        # where "short" means effectively instantaneous file creation / copy
        # / touch.
        #
        # procedure:
        #   - create a file via mkstemp
        #   - create file monitor with detection callback and non-zero settle
        #       time.
        #   - touch file
        #   - check that callback was NOT triggered immediately
        #   - wait settle time / 2, check that cb NOT triggered yet
        #   - wait settle time / 4, check that cb NOT triggered yet
        #   - wait settle time / 4, check that cb HAS been called.

        fp = self._mk_test_fp()

        has_triggered = [False]

        def cb(filepath):
            has_triggered[0] = True
            self.assertEqual(filepath, fp)

        interval = 0.01
        settle = 0.1
        monitor = smqtk.utils.file_utils.FileModificationMonitor(fp, interval,
                                                                 settle, cb)
        try:
            monitor.start()
            # file not touched, should still be waiting
            self.assertEqual(monitor.state, monitor.STATE_WAITING)
            self.assertFalse(has_triggered[0])

            time.sleep(interval)
            smqtk.utils.file_utils.touch(fp)
            time.sleep(interval*2)
            monitor._log.info('checking')
            self.assertFalse(has_triggered[0])
            self.assertEqual(monitor.state, monitor.STATE_WATCHING)

            time.sleep(settle / 2.)
            monitor._log.info('checking')
            self.assertEqual(monitor.state, monitor.STATE_WATCHING)
            self.assertFalse(has_triggered[0])

            time.sleep(settle / 4.)
            monitor._log.info('checking')
            self.assertEqual(monitor.state, monitor.STATE_WATCHING)
            self.assertFalse(has_triggered[0])

            time.sleep(settle / 4.)
            monitor._log.info('checking')
            self.assertTrue(has_triggered[0])

        finally:
            monitor.stop()
            monitor.join()

    def test_long_file_wait(self):
        # procedure:
        #   - create a file via mkstemp
        #   - create file monitor with detection callback and non-zero settle
        #       time.
        #   - setup/start thread that appends to file at an interval that is
        #       less than settle time
        #   - wait and check that cb hasn't been called a few times
        #   - stop appender thread
        #   - check that cb called after settle period

        fp = self._mk_test_fp()

        has_triggered = [False]
        append_interval = 0.02
        monitor_interval = 0.01
        monitor_settle = 0.1

        def cb(filepath):
            has_triggered[0] = True
            self.assertEqual(filepath, fp)

        class AppendThread (threading.Thread):
            def __init__(self):
                super(AppendThread, self).__init__()
                self._s = threading.Event()

            def stop(self):
                self._s.set()

            def stopped(self):
                return self._s.is_set()

            def run(self):
                while not self.stopped():
                    with open(fp, 'a') as f:
                        f.write('0')
                    time.sleep(append_interval)

        m_thread = smqtk.utils.file_utils.FileModificationMonitor(
            fp, monitor_interval, monitor_settle, cb
        )
        a_thread = AppendThread()

        try:
            m_thread.start()
            a_thread.start()

            time.sleep(monitor_settle)
            m_thread._log.info('checking')
            self.assertFalse(m_thread.stopped())
            self.assertFalse(has_triggered[0])
            self.assertEqual(m_thread.state, m_thread.STATE_WATCHING)

            time.sleep(monitor_settle)
            m_thread._log.info('checking')
            self.assertFalse(m_thread.stopped())
            self.assertFalse(has_triggered[0])
            self.assertEqual(m_thread.state, m_thread.STATE_WATCHING)

            a_thread.stop()

            time.sleep(monitor_settle)
            m_thread._log.info('checking')
            self.assertTrue(has_triggered[0])

        finally:
            a_thread.stop()
            m_thread.stop()

    def test_invalid_params(self):
        fp = self._mk_test_fp()

        # Invalid path value
        self.assertRaises(
            ValueError,
            smqtk.utils.file_utils.FileModificationMonitor,
            '/not/real', 1, 1, lambda p: None
        )
        # Invalid timers values
        self.assertRaises(
            ValueError,
            smqtk.utils.file_utils.FileModificationMonitor,
            fp, -1, 1, lambda p: None
        )
        self.assertRaises(
            ValueError,
            smqtk.utils.file_utils.FileModificationMonitor,
            fp, 1, -1, lambda p: None
        )



from __future__ import division, print_function
import unittest

import mock

from smqtk.algorithms.relevancy_index import RelevancyIndex


class DummyRI (RelevancyIndex):

    @classmethod
    def is_usable(cls):
        return True

    def rank(self, pos, neg):
        pass

    def get_config(self):
        pass

    def count(self):
        return 0

    def build_index(self, descriptors):
        pass


class TestSimilarityIndexAbstract (unittest.TestCase):

    def test_count(self):
        index = DummyRI()
        self.assertEqual(index.count(), 0)
        self.assertEqual(index.count(), len(index))

        # Pretend that there were things in there. Len should pass it though
        index.count = mock.Mock()
        index.count.return_value = 5
        self.assertEqual(len(index), 5)



from __future__ import division, print_function
import unittest

import mock

from smqtk.algorithms.nn_index.lsh.functors import \
    LshFunctor, get_lsh_functor_impls


class TestLshFunctorImplGetter (unittest.TestCase):

    @mock.patch('smqtk.algorithms.nn_index.lsh.functors.plugin.get_plugins')
    def test_get_lsh_functor_impls_no_reload(self, m_get_plugins):
        get_lsh_functor_impls()
        m_get_plugins.assert_called_once()
        self.assertEqual(m_get_plugins.call_args[0][0],
                         'smqtk.algorithms.nn_index.lsh.functors')
        self.assertEqual(m_get_plugins.call_args[0][2],
                         'LSH_FUNCTOR_PATH')
        self.assertEqual(m_get_plugins.call_args[0][3],
                         'LSH_FUNCTOR_CLASS')
        self.assertEqual(m_get_plugins.call_args[0][4],
                         LshFunctor)
        self.assertFalse(m_get_plugins.call_args[1]['reload_modules'])

    @mock.patch('smqtk.algorithms.nn_index.lsh.functors.plugin.get_plugins')
    def test_get_lsh_functor_impls_with_reload(self, m_get_plugins):
        get_lsh_functor_impls(True)
        m_get_plugins.assert_called_once()
        self.assertEqual(m_get_plugins.call_args[0][0],
                         'smqtk.algorithms.nn_index.lsh.functors')
        # m_get_plugins.call_args[0][1] is a path depending on where the python
        # code is.
        self.assertEqual(m_get_plugins.call_args[0][2],
                         'LSH_FUNCTOR_PATH')
        self.assertEqual(m_get_plugins.call_args[0][3],
                         'LSH_FUNCTOR_CLASS')
        self.assertEqual(m_get_plugins.call_args[0][4],
                         LshFunctor)
        self.assertTrue(m_get_plugins.call_args[1]['reload_modules'])



from collections import OrderedDict

import threading
import time
import unittest

from smqtk.utils.read_write_lock import \
    ContextualReadWriteLock


def wait_for_value(f, timeout):
    """
    Wait a specified timeout period of time (seconds) for the given
    function to execute successfully.

    `f` usually wraps an assertion function.

    :param f: Assertion function.
    :type f: () -> None

    :param timeout: Time out in seconds to wait for convergence.
    :type timeout: float
    """
    s = time.time()
    neq = True
    while neq:
        try:
            f()
            # function passed.
            neq = False
        except (Exception, AssertionError):
            # if assertion fails past timeout, actually raise assertion.
            if time.time() - s > timeout:
                raise


class TestContextualReadWriteLock (unittest.TestCase):

    def setUp(self):
        self.state = OrderedDict()

    def wait_for_state(self, k):
        """ Wait forever until a state attribute is True. """
        while k not in self.state or not self.state[k]:
            pass

    # Added asserts

    def assertInState(self, k):
        """ Assert key in state """
        self.assertIn(k, self.state)

    def assertLockFree(self, l):
        self.assertEqual(l._semlock._get_value(), 1)

    def assertLockAcquired(self, l):
        self.assertEqual(l._semlock._get_value(), 0)

    # Unit Tests

    def test_initial_state(self):
        # Test expected lock and value states before use.
        crwl = ContextualReadWriteLock()
        self.assertLockFree(crwl._service_lock)
        self.assertLockFree(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

    def test_read_context_state(self):
        # Test expected state conditions when transitioning into and out of a
        # read-lock context.
        crwl = ContextualReadWriteLock()

        def t1(c):
            with c.read_context():
                self.state['t1_read_acquired'] = True
                self.wait_for_state('t1_release')
            self.state['t1_read_released'] = True

        t1 = threading.Thread(target=t1, args=(crwl,))
        t1.daemon = True
        t1.start()

        # Thread should immediately attempt to acquire read lock.  We should see
        # that it does successfully.
        wait_for_value(lambda: self.assertInState('t1_read_acquired'),
                       1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 1)

        # Trigger thread to release context and check state.
        self.state['t1_release'] = True
        wait_for_value(lambda: self.assertInState('t1_read_released'),
                       1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockFree(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

    def test_write_context_state(self):
        # Test expected state conditions when transitioning into and out of a
        # write-lock context.
        crwl = ContextualReadWriteLock()

        def t1_func(c):
            with c.write_context():
                self.state['t1_write_acquired'] = True
                self.wait_for_state('t1_release')
            self.state['t1_write_released'] = True

        t1 = threading.Thread(target=t1_func, args=(crwl,))
        t1.daemon = True
        t1.start()

        # Thread should immediately attempt to acquire write lock.  We should
        # see that it does successfully.
        wait_for_value(lambda: self.assertInState('t1_write_acquired'),
                       1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

        # Trigger thread to release context and check state.
        self.state['t1_release'] = True
        wait_for_value(lambda: self.assertInState('t1_write_released'),
                       1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockFree(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

    def test_concurrent_read_then_write(self):
        # Test that a thread with a read lock blocks a write lock from entering.
        crwl = ContextualReadWriteLock()

        # Thread 1 function - Read lock
        def t1_func(c):
            with c.read_context():
                self.state['t1_read_acquired'] = True
                self.wait_for_state('t1_release')
            self.state['t1_read_released'] = True

        # Thread 2 function - Write lock
        def t2_func(c):
            self.wait_for_state('t2_acquire')
            with c.write_context():
                self.state['t2_write_acquired'] = True
                self.wait_for_state('t2_release')
            self.state['t2_write_released'] = True

        t1 = threading.Thread(target=t1_func, args=(crwl,))
        t2 = threading.Thread(target=t2_func, args=(crwl,))
        t1.daemon = t2.daemon = True
        t1.start()
        t2.start()

        # Upon starting threads, t1 should get read lock and t2 should not have
        # done anything yet.
        wait_for_value(lambda: self.assertInState('t1_read_acquired'), 1.0)
        self.assertNotIn('t2_write_acquired', self.state)
        self.assertLockFree(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 1)

        # t2 should attempt to acquire write context but be blocked.  We should
        # see that the service lock is acquired and that 't2_write_acquired' is
        # not set.
        self.state['t2_acquire'] = True
        wait_for_value(lambda: self.assertLockAcquired(crwl._service_lock), 1.0)
        self.assertNotIn('t2_write_acquired', self.state)
        self.assertLockAcquired(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 1)

        # Releasing t1's read lock should cause t2 to acquire write lock.
        self.state['t1_release'] = True
        wait_for_value(lambda: self.assertInState('t1_read_released'), 1.0)
        wait_for_value(lambda: self.assertInState('t2_write_acquired'), 1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

        # t2 should now be able to release the write lock like normal
        self.state['t2_release'] = True
        wait_for_value(lambda: self.assertInState('t2_write_released'), 1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockFree(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

    def test_concurrent_write_then_read(self):
        # Test that a thread with a read lock blocks a write lock from entering.
        crwl = ContextualReadWriteLock()

        # Thread 1 function - Write lock
        def t1_func(c):
            with c.write_context():
                self.state['t1_write_acquired'] = True
                self.wait_for_state('t1_release')
            self.state['t1_write_released'] = True

        # Thread 2 function - Read lock
        def t2_func(c):
            self.wait_for_state('t2_acquire')
            self.state['t2_read_attempt'] = True
            with c.read_context():
                self.state['t2_read_acquired'] = True
                self.wait_for_state('t2_release')
            self.state['t2_read_released'] = True

        t1 = threading.Thread(target=t1_func, args=(crwl,))
        t2 = threading.Thread(target=t2_func, args=(crwl,))
        t1.daemon = t2.daemon = True
        t1.start()
        t2.start()

        # Upon starting threads, t1 should get write lock and t2 should not have
        # done anything yet.
        wait_for_value(lambda: self.assertInState('t1_write_acquired'), 1.0)
        self.assertNotIn('t2_read_acquired', self.state)
        self.assertLockFree(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

        # t2 should attempt to acquire read context but be blocked.  We should
        # see that the service lock is acquired and that 't2_read_acquired' is
        # not set.
        self.state['t2_acquire'] = True
        wait_for_value(lambda: self.assertLockAcquired(crwl._service_lock), 1.0)
        self.assertNotIn('t2_write_acquired', self.state)
        self.assertLockAcquired(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockAcquired(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)

        # Releasing t1's write lock should cause t2 to acquire read lock.
        self.state['t1_release'] = True
        wait_for_value(lambda: self.assertInState('t1_write_released'), 1.0)
        wait_for_value(lambda: self.assertInState('t2_read_acquired'), 1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockAcquired(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 1)

        # t2 should now be able to release the read lock like normal
        self.state['t2_release'] = True
        wait_for_value(lambda: self.assertInState('t2_read_released'), 1.0)
        self.assertLockFree(crwl._service_lock)
        self.assertLockFree(crwl._resource_lock)
        self.assertLockFree(crwl._reader_count_lock)
        self.assertEqual(crwl._reader_count, 0)
